//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
//  Copyright (c) 2005 Michel Zandstra [michelzandstra@gmail.com]
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 
#include <map>
#include <vector>
#include <cstdlib>

#include <boost/config.hpp>
#include <boost/assert.hpp>

#if defined(BOOST_HAS_UNISTD_H)
#include <unistd.h>
#endif

#include <saga/util.hpp>
#include <saga/exception.hpp>
#include <saga/url.hpp>
#include <saga/adaptors/task.hpp>

#include <impl/config.hpp>
#include <impl/engine/task_get_bulk_adaptor.hpp>

#include "default_file.hpp"
#include "default_namespace_entry_impl.hpp"
#include "common_helpers.hpp"

#ifndef  MAX_PATH
# define MAX_PATH _POSIX_PATH_MAX
#endif

/*
 * constructor
 */
file_cpi_impl::file_cpi_impl (proxy* p, cpi_info const& info, 
        saga::ini::ini const& glob_ini, saga::ini::ini const& adap_ini, 
        TR1::shared_ptr<saga::adaptor> adaptor)
:   base_type (p, info, glob_ini, adap_ini, adaptor, cpi::Noflags) 
{
  // gets properly initialized
  mutex_type::scoped_lock lock(mtx_);
  if (file_.is_open ()) 
      return;   // this file has been opened already

  // first usage of this adaptor
  instance_data data (this);
  saga::url file_url(data->location_);
  std::string url = file_url.get_path();

  std::string host(file_url.get_host());
  if (!host.empty() && host != "localhost" && !boost::iequals(host, ::detail::get_hostname()))
  {
    SAGA_OSSTREAM strm;
    strm << "file_cpi_impl::init: cannot handle remote file: " << url;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
  }

  std::string scheme(file_url.get_scheme());
  if (!scheme.empty() && scheme != "file" && scheme != "any")
  {
    SAGA_OSSTREAM strm;
    strm << "file_cpi_impl::init: cannot handle remote file: " << url;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
  }

  saga::filesystem::flags mode = (saga::filesystem::flags)data->mode_;
  
  namespace fs = boost::filesystem;
  fs::path fpath (url, fs::native);
  if ( (mode & saga::filesystem::Create) && (mode & saga::filesystem::Exclusive) ) 
  {

    if ( fs::exists (fpath) ) 
    {
      SAGA_OSSTREAM strm;
      strm << "file_cpi_impl::init: file already exists: " << url;
      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::AlreadyExists);
    }
  }

  // Create file if saga::filesystem::Create flag is given and file does not exist
  if ( mode & saga::filesystem::Create && !fs::exists(fpath) )
  {
    file_.open (url.c_str (), std::ios::out);
    if (!file_.is_open())
    {
      SAGA_OSSTREAM strm;
      strm << "file_cpi_impl::init: could not create file: " << url;
      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::DoesNotExist);
    }
    file_.close();

    data->mode_ &= ~saga::filesystem::Create;   // no need to create this file twice
  }

  file_.open (url.c_str (), ::detail::translate_open_mode (mode));
  if ( ! file_.is_open () )
  { 
    SAGA_OSSTREAM strm;
    strm << "file_cpi_impl::init: could not open file: " << url;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::DoesNotExist);
  }
  
  // position the pointer to the given position if appropriate
  if (0 != data->pointer_) 
  {
    file_.seekg (data->pointer_, std::ios::beg);
  }
  else if (mode & saga::filesystem::Append)
  {
  // position the file pointer to the end, if the file was opened in Append mode
    file_.seekg (0, std::ios::end);
    data->pointer_ = file_.tellg();
  }
}

/*
 * destructor
 */
file_cpi_impl::~file_cpi_impl (void)
{
    try {
        mutex_type::scoped_lock lock(mtx_);
        if (file_.is_open())
            file_.close();
    }
    catch (...) {
        ;
    }
}

/*
 * SAGA API functions
 */
void file_cpi_impl::sync_get_size (saga::off_t& size_out)
{
    // TODO: consider to use stat (2)
    mutex_type::scoped_lock lock(mtx_);

    BOOST_ASSERT(file_.is_open());
    
    // position to the end of the file
    if (file_.eof())
        file_.clear(std::ios::eofbit);
    file_.seekg (0, std::ios::end);
    size_out = file_.tellg ();      // figure out the file size

    // reposition file pointer
    instance_data data (this);
    file_.seekg (data->pointer_, std::ios::beg);
}

void file_cpi_impl::sync_close (saga::void_t & ret, double)
{
    mutex_type::scoped_lock lock(mtx_);
    if (file_.is_open())
        file_.close();
}

void file_cpi_impl::sync_remove (saga::void_t & ret, int flags)
{
    // first make sure this file is closed 
    {
        mutex_type::scoped_lock lock(mtx_);
        if (file_.is_open())
            file_.close();
    }
    
    // remove the file
    base_type::sync_remove(ret, flags);
}

void file_cpi_impl::sync_read (saga::ssize_t & len_out, 
    saga::mutable_buffer data, saga::ssize_t len_in)
{
  // validate parameters
  if (len_in < 0)
  {
      SAGA_ADAPTOR_THROW(
          "file_cpi_impl::sync_read: 'len_in' should not be negative", 
          saga::BadParameter);
  }
  if (data.get_size() != -1 && len_in > data.get_size())
  {
      SAGA_ADAPTOR_THROW(
          "file_cpi_impl::sync_read: buffer too small for data to read", 
          saga::BadParameter);
  }

  // initialize buffer
  if (-1 == data.get_size())
  {
      data.set_size(len_in);    // use implementation managed memory
  }
  
  // calculate length of rest of the file
  mutex_type::scoped_lock lock(mtx_);

  BOOST_ASSERT(file_.is_open());

  file_.seekg (0, std::ios::end);

  std::fstream::pos_type length = file_.tellg ();
  {
    instance_data data (this);
    file_.seekg (data->pointer_, std::ios::beg);
  }

  length -= file_.tellg ();

  // set length to read 
  if ( (saga::ssize_t) length < data.get_size() )
  {
    len_out = length;
  }
  else
  {
    len_out = data.get_size();
  }

  // ensure, that the stream operates on bytes
  BOOST_ASSERT (sizeof (std::fstream::char_type) == sizeof (saga::char_t));

  /* read data */
  file_.read ((char *)data.get_data(), (std::streamsize)data.get_size());
  if (file_.rdstate () & (std::ios::badbit | std::ios::failbit | std::ios::eofbit))
  {
    // wasn't able to read the required amount of data
    if (file_.rdstate () & std::ios::eofbit)
    {
      len_out = file_.gcount ();
    }
    else
    {
      SAGA_OSSTREAM strm;
      instance_data data (this);
      strm << "file_cpi_impl::read: could not read from file: " << data->location_;

      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::Unexpected);
    }
  }

  /* set buffer and file pointer */
  {
    instance_data data (this);
    data->pointer_ += (saga::off_t) len_out;
  }
}


void file_cpi_impl::sync_write (saga::ssize_t& len_out, 
    saga::const_buffer data, saga::ssize_t len_in)
{
  // validate parameters
  if (len_in < 0)
  {
      SAGA_ADAPTOR_THROW(
          "file_cpi_impl::sync_write: 'len_in' should not be negative", 
          saga::BadParameter);
  }

  // write data 
  mutex_type::scoped_lock lock(mtx_);

  BOOST_ASSERT(file_.is_open());
  
  {
    instance_data data (this);
    file_.seekp (data->pointer_, std::ios::beg);
  }

  // ensure, that the stream operates on bytes
  BOOST_ASSERT (sizeof (std::fstream::char_type) == sizeof (saga::char_t));

  file_.write ((char const*)data.get_data(), (std::streamsize)data.get_size());
  if ( file_.rdstate () & (std::ios::badbit | std::ios::failbit) )
  {
    // wasn't able to write the required amount of data
    SAGA_OSSTREAM strm;
    instance_data data (this);
    strm << "file_cpi_impl::write: could not write to file: " << data->location_;

    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::Unexpected);
  }

  {
    instance_data instdata (this);

    len_out             = data.get_size();
    instdata->pointer_ += (saga::off_t)   len_out;
  }
}

void file_cpi_impl::sync_seek (saga::off_t & out, 
                               saga::off_t   offset, 
                               saga::filesystem::seek_mode whence)
{
  /* set position of get pointer */
  mutex_type::scoped_lock lock(mtx_);

  BOOST_ASSERT(file_.is_open());
  
  switch ( (saga::filesystem::seek_mode) whence ) 
  {
    case saga::filesystem::Start:   file_.seekg (offset, std::ios::beg); break;
    case saga::filesystem::Current: file_.seekg (offset, std::ios::cur); break;
    case saga::filesystem::End:     file_.seekg (offset, std::ios::end); break;
    default:
    {
      // wasn't able to seek 
      SAGA_OSSTREAM strm;
      strm << "file_cpi_impl::seek: bogus seek mode: " << (int) whence;
      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::Unexpected);
    }
    break;
  }

  {
    instance_data data (this);
    data->pointer_ = file_.tellg ();    
    out = data->pointer_;
  }
}

#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS)
///////////////////////////////////////////////////////////////////////////////
// This adaptor implements the async functions based on its own synchronous
// functions for testing purposes only. Since there is no principal need
// to do so, we allow these to be 'switched off'.

saga::task_type file_cpi_impl::async_get_size (saga::off_t& size_out)
{
    return (saga::adaptors::task("file_cpi_impl::sync_get_size",
        shared_from_this(), &file_cpi_impl::sync_get_size, 
        TR1::ref(size_out)));
}

saga::task_type file_cpi_impl::async_close (saga::void_t & ret, double timeout)
{
    return (saga::adaptors::task("file_cpi_impl::sync_close",
        shared_from_this(), &file_cpi_impl::sync_close, TR1::ref(ret),
        timeout));
}

saga::task_type file_cpi_impl::async_remove (saga::void_t & ret, int flags)
{
    return (saga::adaptors::task("file_cpi_impl::sync_remove",
        shared_from_this(), &file_cpi_impl::sync_remove, TR1::ref(ret),
        flags));
}

saga::task_type 
file_cpi_impl::async_read (saga::ssize_t & len_out,
    saga::mutable_buffer data, saga::ssize_t len_in)
{
    return (saga::adaptors::task("file_cpi_impl::sync_read",
        shared_from_this(), &file_cpi_impl::sync_read, TR1::ref(len_out), 
        data, len_in));
}

saga::task_type 
file_cpi_impl::async_write (saga::ssize_t& len_out, 
    saga::const_buffer data, saga::ssize_t len_in)
{
    return (saga::adaptors::task("file_cpi_impl::sync_write",
        shared_from_this(), &file_cpi_impl::sync_write, TR1::ref(len_out), 
        data, len_in));
}

saga::task_type file_cpi_impl::async_seek(saga::off_t & out, saga::off_t offset, 
    saga::filesystem::seek_mode whence) 
{
    return (saga::adaptors::task("file_cpi_impl::sync_seek",
        shared_from_this(), &file_cpi_impl::sync_seek, TR1::ref(out),  
        offset, whence));
}

#endif // SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS

