//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
//  Copyright (c) 2006      Stephan Hirmer (shirmer@cct.lsu.edu)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef ADAPTORS_DEFAULT_FILE_DEFAULT_NAME_SPACE_DIR_IMPL_HPP
#define ADAPTORS_DEFAULT_FILE_DEFAULT_NAME_SPACE_DIR_IMPL_HPP

#include <map>
#include <vector>

#include <saga/util.hpp>
#include <saga/exception.hpp>
#include <saga/url.hpp>
#include <impl/config.hpp>

#include <saga/adaptors/bulk_container.hpp>
#include <saga/uuid.hpp>

#include "default_namespace_entry.hpp"
#include "default_namespace_dir.hpp"
#include "common_helpers.hpp"

#include "default_namespace_entry_impl.hpp"

#include <sys/types.h>
#include <sys/stat.h>

#if defined(BOOST_HAS_UNISTD_H) 
#include <unistd.h>
#endif

#ifndef  MAX_PATH
# define MAX_PATH _POSIX_PATH_MAX
#endif

/*
 * constructor
 */
template <typename Base>
inline
namespace_dir_cpi_impl<Base>::namespace_dir_cpi_impl (saga::impl::proxy* p, 
        saga::impl::v1_0::cpi_info const& info, saga::ini::ini const& glob_ini, 
        saga::ini::ini const& adap_ini, TR1::shared_ptr<saga::adaptor> adaptor,
        saga::impl::v1_0::cpi::flags flags)
:   base_cpi(p, info, glob_ini, adap_ini, adaptor, flags)
{
}

/*
 * destructor
 */
template <typename Base>
inline
namespace_dir_cpi_impl<Base>::~namespace_dir_cpi_impl (void)
{
}

/*
 * SAGA API functions
 */

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_exists (bool & exists, std::string url)
{
  namespace fs = boost::filesystem;

  // verify this file can be handled
  if (!::detail::file_islocal(url)) 
  {
    SAGA_OSSTREAM strm;
    strm << "namespace_dir_cpi_impl<Base>::sync_exists: "
        "cannot handle remote target directory: " << url;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
  }

  instance_data data (this);

  fs::path name (::detail::get_filepath(url), fs::native);
  fs::path path (::detail::get_filepath(data->location_), fs::native);

  if ( ! name.has_root_path () )
    path /= name;
  else
    path = name;

  exists = fs::exists (path);
}

template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_is_dir (bool & is_dir, std::string url, 
    int flags)
{
  namespace fs = boost::filesystem;

  // verify the file can be handled
  if (!::detail::file_islocal(url)) 
  {
    SAGA_OSSTREAM strm;
    strm << "namespace_dir_cpi_impl<Base>::sync_is_dir: "
        "cannot handle remote target entry: " << url;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
  }

  instance_data data (this);

  fs::path name (::detail::get_filepath(url), fs::native);
  fs::path path (::detail::get_filepath(data->location_), fs::native);

  if ( ! name.has_root_path () )
    path /= name;
  else
    path = name;

  is_dir = fs::exists (path) && fs::is_directory (path);
}

template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_is_entry (bool & is_entry, std::string url,
    int flags)
{
  namespace fs = boost::filesystem;

  // verify the file can be handled
  if (!::detail::file_islocal(url)) 
  {
    SAGA_OSSTREAM strm;
    strm << "namespace_dir_cpi_impl<Base>::sync_is_entry: "
        "cannot handle remote target entry: " << url;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
  }

  instance_data data (this);

  fs::path name (::detail::get_filepath(url), fs::native);
  fs::path path (::detail::get_filepath(data->location_), fs::native);

  if ( ! name.has_root_path () )
    path /= name;
  else
    path = name;

  is_entry = fs::exists (path) && ! fs::is_directory (path);
}

template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_is_link (bool & is_link, std::string url, 
    int flags)
{
  namespace fs = boost::filesystem;

  instance_data data (this);

  // verify the file can be handled
  if (!::detail::file_islocal(url)) 
  {
    SAGA_OSSTREAM strm;
    strm << "namespace_dir_cpi_impl<Base>::sync_is_link: "
        "cannot handle remote target entry: " << url;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
  }

  fs::path name (::detail::get_filepath(url), fs::native);
  fs::path path (::detail::get_filepath(data->location_), fs::native);

  if ( ! name.has_root_path () )
    path /= name;
  else
    path = name;

#if BOOST_VERSION >= 103400
  is_link = fs::exists (path) && fs::symbolic_link_exists (url);
#elif defined(linux) || defined(__linux) || defined(__linux__) || defined(__APPLE__) 
  {
    is_link = false;
    struct stat out;
    lstat (path.string ().c_str (), &out);

    is_link = S_ISLNK(out.st_mode) ? true : false;
  }
#else
  SAGA_ADAPTOR_THROW("Not implemented.", saga::NotImplemented);
#endif
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_change_dir (saga::void_t&, std::string new_dir)
{
    namespace fs = boost::filesystem;

    // verify the file can be handled
    if (!::detail::file_islocal(new_dir)) 
    {
        SAGA_OSSTREAM strm;
        strm << "namespace_dir_cpi_impl<Base>::sync_change_dir: "
            "cannot handle remote target directory: " << new_dir;
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
    }

    instance_data data (this);

    fs::path name (::detail::get_filepath(new_dir),         fs::native);
    fs::path path (::detail::get_filepath(data->location_), fs::native);

    if ( ! name.has_root_path () )
        path /= name;
    else
        path = name;
    path.normalize();
    
    if (!(fs::exists(path) && fs::is_directory (path)))
    {
        SAGA_ADAPTOR_THROW(
            path.string () + ": doesn't refer to a directory object",
            saga::DoesNotExist);
    }

    saga::url u(data->location_);
    u.set_path(path.string());
    data->location_ = u.get_url();
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_list(std::vector <std::string> & retval, 
    std::string pattern, int flags)
{
    namespace fs = boost::filesystem;

    instance_data data (this);
    fs::path path (::detail::get_filepath(data->location_), fs::native);
    std::string host(::detail::get_scheme_authority(data->location_));
    
    fs::directory_iterator dir_end = fs::directory_iterator();
    for (fs::directory_iterator dir_it(path); dir_it != dir_end; ++dir_it)
    {
        //retval.push_back(host + (*dir_it).leaf()); see --> Spec. p. 141
        retval.push_back((*dir_it).leaf());
    }
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_find(std::vector <std::string> & retval, 
    std::string entry, int)
{
    SAGA_ADAPTOR_THROW(
        "namespace_dir_cpi_impl<Base>::sync_find not implemented yet", 
        saga::NotImplemented);
}

namespace {

    // copy a directory, even recursively, if appropriate
    void 
    copy_directory(boost::filesystem::path src, boost::filesystem::path dst, 
        int open_mode)
    {
        namespace fs = boost::filesystem;
        
        // ensure the whole destination paths exists, if appropriate
        if (open_mode & saga::name_space::CreateParents)
            fs::create_directories(dst);

        // copy all entries
        fs::directory_iterator dir_end = fs::directory_iterator();
        for (fs::directory_iterator dir_it(src); dir_it != dir_end; ++dir_it)
        {
            // construct the destination path
            fs::path curr_dst (dst.branch_path());
            curr_dst /= (*dir_it).leaf();
            
            // if the destination is a directory and we need to work recursively
            if (fs::is_directory(curr_dst)) {
                if (open_mode & saga::name_space::Recursive)
                    copy_directory(*dir_it, curr_dst, open_mode);
            }
            else {
                // copy the file
                if ((open_mode & saga::name_space::Overwrite) && fs::exists(curr_dst))
                    fs::remove(curr_dst);

                fs::copy_file(*dir_it, curr_dst);
            }
        }        
    }
}   // namespace

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_copy(saga::void_t & ret, 
                                          std::string source, 
                                          std::string destination,
                                          int open_mode)
{
    namespace fs = boost::filesystem;
    
    instance_data data (this);
        
    // verify the files can be handled
    if (!::detail::file_islocal(source)) 
    {
      SAGA_OSSTREAM strm;
      strm << "namespace_dir_cpi_impl<Base>::sync_copy: "
          "cannot handle remote source file: " << source;
      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
    }

    if (!::detail::file_islocal(destination)) 
    {
      SAGA_OSSTREAM strm;
      strm << "namespace_dir_cpi_impl<Base>::sync_copy: "
          "cannot handle remote destination file: " << destination;
      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
    }

    // handle the files
    fs::path src_location (::detail::get_filepath(data->location_), fs::native);
    fs::path dst_location (src_location);
    
    // complete paths
    fs::path src_path  (::detail::get_filepath(source), fs::native);
    fs::path dest_path (::detail::get_filepath(destination), fs::native);

    if ( ! src_path.has_root_path () )
        src_location /= src_path;
    else
        src_location = src_path;

    if ( ! dest_path.has_root_path () )
        dst_location /= dest_path;
    else
        dst_location = dest_path;

    try {
        if (fs::is_directory(src_location)) {
            // src location refers to a is a directory
            copy_directory(src_location, dst_location, open_mode);
        }
        else {
            // remove the file if it exists and we should overwrite
            if ((open_mode & saga::name_space::Overwrite) && fs::exists(dst_location))
                fs::remove(dst_location);

            fs::copy_file(src_location, dst_location);
        }
    }
    catch(fs::filesystem_error const& e) {
        SAGA_ADAPTOR_THROW(e.what(), saga::NoSuccess);
    }
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_move(saga::void_t & ret, 
                                            std::string source, 
                                            std::string destination, 
                                            int flags)
{
    this->sync_copy(ret, source, destination, flags);
    this->sync_remove(ret, source, flags);   
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_remove(saga::void_t & ret, std::string url, 
                                          int flags)
{
    namespace fs = boost::filesystem;
       
    // verify the file can be handled
    if (!::detail::file_islocal(url)) 
    {
        SAGA_OSSTREAM strm;
        strm << "namespace_dir_cpi_impl<Base>::sync_remove: "
            "cannot handle remote target file: " << url;
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
    }

    instance_data data (this);
    fs::path src_location (::detail::get_filepath(data->location_), fs::native);
    
    // complete paths
    fs::path src_path  (::detail::get_filepath(url), fs::native);
   
    if ( ! src_path.has_root_path () )    
        src_location /= src_path;
    else
        src_location = src_path;
    
    if (!fs::exists(src_location)) {
        SAGA_ADAPTOR_THROW("directory::remove: Can't remove directory: "
          "Does not exist", saga::DoesNotExist);
    }
    
    if (fs::is_directory(src_location)) 
    {
        if ( (saga::name_space::Recursive != flags) && !(fs::is_empty(src_location)) )
        {
            SAGA_ADAPTOR_THROW("directory::remove: Can't remove directory. "
                "Please use recursive mode!", saga::BadParameter);
        }
        else
        {    
            remove_all(src_location);
        }
    }
    else 
         fs::remove(src_location);
}

template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_make_dir(saga::void_t & ret, std::string url, 
    int open_mode)
{
    namespace fs = boost::filesystem;
       
    // verify the file can be handled
    if (!::detail::file_islocal(url)) 
    {
        SAGA_OSSTREAM strm;
        strm << "namespace_dir_cpi_impl<Base>::sync_make_dir: "
            "cannot handle remote target directory: " << url;
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
    }

    instance_data data (this);
    fs::path src_location (::detail::get_filepath(data->location_), fs::native);
    
    // complete paths
    fs::path src_path  (::detail::get_filepath(url), fs::native);
   
    if ( ! src_path.has_root_path () )    
        src_location /= src_path;
    else
        src_location = src_path;
    try
    {
        fs::create_directory(src_location);
    }
    catch(fs::filesystem_error const& e) 
    {
        SAGA_ADAPTOR_THROW(e.what(), saga::NoSuccess);
    }
}


template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_link(saga::void_t & ret, 
                                        std::string source, 
                                        std::string url, 
                                        int open_mode)
{
    SAGA_ADAPTOR_THROW("namespace_dir_cpi_impl<Base>::sync_link not yet implemented", 
                       saga::NotImplemented);
}

template <typename Base>
inline void 
namespace_dir_cpi_impl<Base>::sync_read_link(std::string & ret, 
  std::string source)
{
  namespace fs = boost::filesystem;

  // verify the file can be handled
  if (!::detail::file_islocal(source)) 
  {
    SAGA_OSSTREAM strm;
    strm << "namespace_dir_cpi_impl<Base>::sync_read_link: "
        "cannot handle remote target file: " << source;
    SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
  }

  instance_data data (this);

  fs::path name (::detail::get_filepath(source), fs::native);
  fs::path path (::detail::get_filepath(data->location_), fs::native);

  if ( ! name.has_root_path () )
    path /= name;
  else
    path = name;

#if defined(linux) || defined(__linux) || defined(__linux__) || defined(__APPLE__)
  {
    char* buf = (char*) calloc (sizeof (char), MAX_PATH + 1);
    readlink (path.string ().c_str (), buf, MAX_PATH);
    ret = buf;
    free (buf);
  }
#else
    SAGA_ADAPTOR_THROW("namespace_dir_cpi_impl<Base>::sync_read_link not yet implemented",
                       saga::NotImplemented);
#endif
}

#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS)
///////////////////////////////////////////////////////////////////////////////
// This adaptor implements the async functions based on its own synchronous
// functions for testing purposes only. SInce there is no principal need
// to do so, we allow these to be 'switched off'.

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_exists (bool & exists, std::string url)
{
    return (saga::adaptors::task("namespace_dir_cpi::sync_exists",
                                    this->shared_from_this(), 
                                    &namespace_dir_cpi_impl<Base>::sync_exists,
                                    TR1::ref (exists), url ));
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_is_dir (bool & is_dir, std::string url, 
    int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::sync_is_dir",
                                    this->shared_from_this(), 
                                    &namespace_dir_cpi_impl<Base>::sync_is_dir,
                                    TR1::ref (is_dir), url, flags));
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_is_entry (bool & is_entry, std::string url,
    int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::sync_is_entry",
                                    this->shared_from_this(), 
                                    &namespace_dir_cpi_impl<Base>::sync_is_entry,
                                    TR1::ref (is_entry), url, flags));
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_is_link (bool & is_link, std::string url,
    int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::sync_is_link",
                                    this->shared_from_this(), 
                                    &namespace_dir_cpi_impl<Base>::sync_is_link,
                                    TR1::ref (is_link), url, flags));
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_change_dir (saga::void_t & retval, 
    std::string new_dir)
{
    return (saga::adaptors::task("namespace_dir_cpi::sync_change_dir",
                                    this->shared_from_this(), 
                                    &namespace_dir_cpi_impl<Base>::sync_change_dir,
                                    TR1::ref (retval), new_dir));
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_copy(saga::void_t & ret, 
                                           std::string source, 
                                           std::string destination,
                                           int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::copy",
                                 this->shared_from_this(), 
                                 &namespace_dir_cpi_impl<Base>::sync_copy,
                                 TR1::ref (ret),
                                 source,
                                 destination,
                                 flags
#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS)
                               , &namespace_dir_cpi_impl<Base>::prepare_copy 
#endif
                             )); 
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_move(saga::void_t & ret, 
                                            std::string source, 
                                            std::string destination, 
                                            int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::move",
                                 this->shared_from_this(), 
                                 &namespace_dir_cpi_impl<Base>::sync_move,
                                 TR1::ref (ret),
                                 source,
                                 destination,
                                 flags
#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS)
                               , &namespace_dir_cpi_impl<Base>::prepare_move 
#endif
                            )); 
}   

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_remove(saga::void_t & ret, 
                                                 std::string url, 
                                                 int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::remove",
                                 this->shared_from_this(), 
                                 &namespace_dir_cpi_impl<Base>::sync_remove,
                                 TR1::ref (ret),
                                 url,
                                 flags
#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS)
                               , &namespace_dir_cpi_impl<Base>::prepare_remove 
#endif
                            ));
}      

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_make_dir(saga::void_t & ret, 
                                                   std::string url, 
                                                   int open_mode)
{
    return (saga::adaptors::task("namespace_dir_cpi::make_dir",
                                 this->shared_from_this(), 
                                 &namespace_dir_cpi_impl<Base>::sync_make_dir,
                                 TR1::ref (ret),
                                 url,
                                 open_mode )); 
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_link(saga::void_t & ret, 
                                               std::string source, 
                                               std::string url, 
                                               int open_mode)
{
    return (saga::adaptors::task("namespace_dir_cpi::link",
                                 this->shared_from_this(), 
                                 &namespace_dir_cpi_impl<Base>::sync_link,
                                 TR1::ref (ret),
                                 source,
                                 url,
                                 open_mode )); 
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_read_link(std::string & ret, 
                                                    std::string source)
{
    return (saga::adaptors::task("namespace_dir_cpi::read_link",
                                 this->shared_from_this(), 
                                 &namespace_dir_cpi_impl<Base>::sync_read_link,
                                 TR1::ref (ret),
                                 source)); 
}
                                            
template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_list(
    std::vector <std::string> & retval, std::string pattern, int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::list",
        this->shared_from_this(), &namespace_dir_cpi_impl<Base>::sync_list,
        TR1::ref (retval), pattern, flags)); 
}

template <typename Base>
inline saga::task_type 
namespace_dir_cpi_impl<Base>::async_find(
    std::vector <std::string> & retval, std::string entry, int flags)
{
    return (saga::adaptors::task("namespace_dir_cpi::find",
        this->shared_from_this(), &namespace_dir_cpi_impl<Base>::sync_find,
        TR1::ref (retval), entry, flags)); 
}
///////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// bulk ops
#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS)
// This adaptor implements the bulk functions for testing purposes only. Since 
// there is no principal need to do so, we allow these to be 'switched off'.
template <typename Base>
inline bool 
namespace_dir_cpi_impl<Base>::prepare_copy(saga::void_t & ret, 
                                           std::string source, 
                                           std::string destination,
                                           int open_mode,
                                           saga::uuid task_uuid/*=NULL*/)
{
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"Reached bulk_prepare for ns_dir::copy"<<std::endl;
    }
    create_bulk_container_or_add_params(this, this->bulk_, 
        &namespace_dir_cpi_impl<Base>::bulk_copy, 
        TR1::ref (ret), source, destination, open_mode, task_uuid );
    return true;
}

template <typename Base>
inline std::vector<saga::uuid> 
namespace_dir_cpi_impl<Base>::bulk_copy(
                      std::vector<TR1::reference_wrapper <saga::void_t> >& rets,
                      std::vector<std::string>& sources, 
                      std::vector<std::string>& dests, 
                      std::vector<int>& modes, std::vector<saga::uuid>& uuids)
{
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"Reached bulk_copy for ns_dir::copy"<<std::endl;
    }
    
    std::vector<TR1::reference_wrapper <saga::void_t> >::iterator ret_it = rets.begin();
    std::vector<std::string>::iterator src_it    = sources.begin();
    std::vector<std::string>::iterator dest_it   = dests.begin();
    std::vector<int>::iterator mod_it    = modes.begin();
    std::vector<saga::uuid>::iterator  id_it     = uuids.begin();
    saga::void_t r;
    
    std::vector<saga::uuid> ret;
    
    while ( ret_it  != rets.end()  && src_it != sources.end() && 
            dest_it != dests.end() && mod_it != modes.end() && id_it != uuids.end() )
    {
        std::string src = (*src_it);
        std::string dst = (*dest_it);
        int         m   = (*mod_it);
        this->sync_copy(r, src, dst, m);
        
        ret.push_back(*id_it);
        
        ++ret_it;
        ++src_it;
        ++dest_it;
        ++mod_it;
        ++id_it;
    }
    return ret;
}

template <typename Base>
inline bool 
namespace_dir_cpi_impl<Base>::prepare_move(saga::void_t & ret, 
                                           std::string source, 
                                           std::string destination,
                                           int open_mode,
                                           saga::uuid task_uuid/*=NULL*/)
{
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"Reached bulk_prepare for ns_dir::move"<<std::endl;
    }
    create_bulk_container_or_add_params(this, this->bulk_, 
        &namespace_dir_cpi_impl<Base>::bulk_move, 
        TR1::ref (ret), source, destination, open_mode, task_uuid );
    return true;
}


template <typename Base>
inline std::vector<saga::uuid> 
namespace_dir_cpi_impl<Base>::bulk_move(
                      std::vector<TR1::reference_wrapper <saga::void_t> >& rets,
                      std::vector<std::string>& sources, 
                      std::vector<std::string>& dests, 
                      std::vector<int>& modes, std::vector<saga::uuid>& uuids)
{
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"Reached bulk_copy for ns_dir::move"<<std::endl;
    }
    
    std::vector<TR1::reference_wrapper <saga::void_t> >::iterator ret_it = rets.begin();
    std::vector<std::string>::iterator src_it    = sources.begin();
    std::vector<std::string>::iterator dest_it   = dests.begin();
    std::vector<int>::iterator mod_it            = modes.begin();
    std::vector<saga::uuid>::iterator  id_it     = uuids.begin();
    saga::void_t r;
    
    std::vector<saga::uuid> ret;
    
    while ( ret_it  != rets.end()  && src_it != sources.end() && 
            dest_it != dests.end() && mod_it != modes.end() && id_it != uuids.end() )
    {
        std::string src = (*src_it);
        std::string dst = (*dest_it);
        int         m   = (*mod_it);
        this->sync_move(r, src, dst, m);
        
        ret.push_back(*id_it);
        
        ++ret_it;
        ++src_it;
        ++dest_it;
        ++mod_it;
        ++id_it;
    }
    return ret;
}


template <typename Base>
inline bool 
namespace_dir_cpi_impl<Base>::prepare_remove(saga::void_t & ret, std::string url, 
                                             int open_mode, 
                                             saga::uuid task_uuid/*=NULL*/)
{
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"Reached bulk_prepare for ns_dir::remove"<<std::endl;
    }
    create_bulk_container_or_add_params(this, this->bulk_, 
        &namespace_dir_cpi_impl<Base>::bulk_remove, TR1::ref (ret), 
        url, open_mode, task_uuid );
    return true;
}

template <typename Base>
inline std::vector<saga::uuid> 
namespace_dir_cpi_impl<Base>::bulk_remove(
                    std::vector<TR1::reference_wrapper <saga::void_t> >& rets,
                    std::vector<std::string>& urls, 
                    std::vector<int>& modes, 
                    std::vector<saga::uuid>& uuids)
{
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"Reached bulk_remove for ns_dir::remove"<<std::endl;
    }

    std::vector<TR1::reference_wrapper <saga::void_t> >::iterator ret_it = rets.begin();
    std::vector<std::string>::iterator url_it    = urls.begin();
    std::vector<int>::iterator         mod_it    = modes.begin();
    std::vector<saga::uuid>::iterator  id_it     = uuids.begin();
    
    saga::void_t r;
    
    std::vector<saga::uuid> ret;
    
    while ( ret_it  != rets.end()  && url_it != urls.end() && 
            mod_it  != modes.end() && id_it != uuids.end() )
    {
        std::string url = (*url_it);
        int         m   = (*mod_it);
        this->sync_remove(r, url,  m);
        
        ret.push_back(*id_it);
        
        ++ret_it;
        ++url_it;
        ++mod_it;
        ++id_it;
    }
    
   return ret;
}

#endif // ! SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS
#endif // ! SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS

#endif // !ADAPTORS_DEFAULT_FILE_DEFAULT_NAME_SPACE_DIR_IMPL_HPP
