//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
//  Copyright (c) 2006 		Ole Weidner (ole.weidner@aei.mpg.de)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(ADAPTORS_DEFAULT_FILE_DEFAULT_NAME_SPACE_ENTRY_IMPL_HPP)
#define ADAPTORS_DEFAULT_FILE_DEFAULT_NAME_SPACE_ENTRY_IMPL_HPP

#include <map>
#include <vector>
#include <stdio.h>
#include <limits.h>

#include <saga/util.hpp>
#include <saga/exception.hpp>
#include <saga/url.hpp>
#include <impl/config.hpp>

#include <saga/packages/namespace/namespace_entry.hpp>

#include <saga/adaptors/bulk_container.hpp>

#include "default_namespace_entry.hpp"
#include "common_helpers.hpp"

#ifndef  MAX_PATH
# define MAX_PATH _POSIX_PATH_MAX
#endif

/*
 * constructor
 */
template <typename Base>
inline 
namespace_entry_cpi_impl<Base>::namespace_entry_cpi_impl (
        saga::impl::proxy                * p,
        saga::impl::v1_0::cpi_info const & info, 
        saga::ini::ini             const & glob_ini, 
        saga::ini::ini             const & adap_ini, 
        TR1::shared_ptr <saga::adaptor>  adaptor,
        saga::impl::v1_0::cpi::flags       flags)
:   base_cpi (p, info, adaptor, flags)
{
}

/*
 * destructor
 */
template <typename Base>
inline 
namespace_entry_cpi_impl<Base>::~namespace_entry_cpi_impl (void)
{
}

/*
 * SAGA API functions
 */

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_exists (bool & exists)
{
    namespace fs = boost::filesystem;

    instance_data data (this);
    fs::path path (::detail::get_filepath(data->location_), fs::native);
    exists = fs::exists (path);
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_is_dir (bool & is_dir, int flags)
{
    namespace fs = boost::filesystem;

    instance_data data (this);
    fs::path path (::detail::get_filepath(data->location_), fs::native);
    is_dir = fs::exists (path) && fs::is_directory (path);
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_is_entry (bool & is_entry, int flags)
{
    namespace fs = boost::filesystem;

    instance_data data (this);
    fs::path path (::detail::get_filepath(data->location_), fs::native);
    is_entry = fs::exists (path) && ! fs::is_directory (path);
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_is_link (bool & is_link, int flags)
{
#if BOOST_VERSION >= 103400
    namespace fs = boost::filesystem;

    instance_data data (this);
    fs::path path (::detail::get_filepath(data->location_), fs::native);
    is_link = fs::exists (path) && fs::symbolic_link_exists (path);
#else
    SAGA_ADAPTOR_THROW("Not implemented.", saga::NotImplemented);
#endif
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_get_url (std::string & url)
{
    saga::url u;
    {
        instance_data data (this);
        u = data->location_;
    }

    // complete url
    if (u.get_scheme().empty())
        u.set_scheme("file");
    if (u.get_host().empty())
        u.set_host("localhost");
        
    url = u.get_url();
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_get_cwd (std::string & url)
{
    namespace fs = boost::filesystem;

    instance_data data (this);
//     saga::url u (::detail::get_filepath(data->location_));
// 
//     fs::path name (u.get_path(), fs::native);
//     if (fs::is_directory (name)) {
//         std::string name_str(name.string());
//         if (name_str.find_last_of(SAGA_PATH_DELIMITERS) == name_str.size()-1) 
//             name = fs::path(name_str.substr(0, name_str.size()-1), fs::native);
//     }

    saga::url result (data->location_);
    
    
    fs::path name (result.get_path(), fs::native);
    
    if( !fs::is_directory(name) ) {
        result.set_path(name.branch_path().string() + "/");
    }
    else {
        
    }

    // complete url
    if (result.get_scheme().empty())
        result.set_scheme("file");
    if (result.get_host().empty())
        result.set_host("localhost");
        
    url = result.get_url();
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_get_name (std::string & url)
{
    namespace fs = boost::filesystem;

    instance_data data (this);
    saga::url u (::detail::get_filepath(data->location_));
    fs::path name (u.get_path(), fs::native);
    
//     if (fs::is_directory (name)) {
//     std::string name_str(name.string());
// 
//         if (name_str.find_last_of(SAGA_PATH_DELIMITERS) == name_str.size()-1) 
//             name = fs::path(name_str.substr(0, name_str.size()-1), fs::native);
//     }
    url = name.leaf();
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_copy (saga::void_t & ret, 
    std::string dest, int flags)
{
    namespace fs = boost::filesystem;

    // verify this file can be handled
    if (!::detail::file_islocal(dest)) 
    {
      SAGA_OSSTREAM strm;
      strm << "namespace_entry_cpi_impl<Base>::sync_copy: "
          "cannot handle remote destination file: " << dest;
      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
    }

    // handle the file
    instance_data data (this);
         
    fs::path src_location (::detail::get_filepath(data->location_), fs::native);
    fs::path dst_location (src_location);
     
    // complete paths
    fs::path dest_path (::detail::get_filepath(dest), fs::native);
 
    if ( ! dest_path.has_root_path () )
        dst_location = dest_path;
    else
        dst_location = dest_path;
          
    try {
        // remove the file if it exists and we should overwrite
        if ((flags & saga::name_space::Overwrite) && fs::exists(dst_location))
            fs::remove(dst_location);
 
        fs::copy_file(src_location, dst_location);
    }
    catch(fs::filesystem_error const& e) {
        SAGA_ADAPTOR_THROW(e.what(), saga::NoSuccess);
    }
}

template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_link (saga::void_t & ret, 
    std::string source, int flags)
{
	SAGA_ADAPTOR_THROW("Not implemented.", saga::NotImplemented);
}

template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_move (saga::void_t & ret, 
    std::string dest, int flags)
{
    namespace fs = boost::filesystem;
     
    // verify this file can be handled
    if (!::detail::file_islocal(dest)) 
    {
      SAGA_OSSTREAM strm;
      strm << "namespace_entry_cpi_impl<Base>::sync_mode: "
          "cannot handle remote destination file: " << dest;
      SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NotImplemented);
    }

    instance_data data (this);
         
    fs::path src_location (::detail::get_filepath(data->location_), fs::native);
    fs::path dst_location (src_location);
     
    // complete paths
    fs::path dest_path (::detail::get_filepath(dest), fs::native);
 
    if ( ! dest_path.has_root_path () )
        dst_location = dest_path;
    else
        dst_location = dest_path;
          
    try {
        // remove the file if it exists and we should overwrite
        if ((flags & saga::name_space::Overwrite) && fs::exists(dst_location))
            fs::remove(dst_location);
 
        fs::copy_file(src_location, dst_location);
        fs::remove(src_location);
    }
    catch(fs::filesystem_error const& e) {
        SAGA_ADAPTOR_THROW(e.what(), saga::NoSuccess);
    }
}

template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_remove (saga::void_t & ret, int flags)
{
     namespace fs = boost::filesystem;
     
     instance_data data (this);
     fs::path src_location (::detail::get_filepath(data->location_), fs::native);

     try {
         fs::remove(src_location);
     }
     catch(fs::filesystem_error const& e) {
         SAGA_ADAPTOR_THROW(e.what(), saga::NoSuccess);
     } 
}

template <typename Base>
inline void 
namespace_entry_cpi_impl<Base>::sync_close (saga::void_t & ret, double timeout)
{
	// Nothing to close here...
}

#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS)
///////////////////////////////////////////////////////////////////////////////
// This adaptor implements the async functions based on its own synchronous
// functions for testing purposes only. Since there is no principal need
// to do so, we allow these to be 'switched off'.

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_get_url (std::string & url) 
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_get_url",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_get_url,
                                 TR1::ref (url) ));
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_get_cwd (std::string & url) 
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_get_cwd",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_get_cwd,
                                 TR1::ref (url) ));
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_get_name (std::string & url) 
{
    return (saga::adaptors::task("namespace_entry_cpi::get_name",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_get_name,
                                 TR1::ref (url)
#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS)
                               , &namespace_entry_cpi_impl<Base>::prepare_get_name
#endif
                             ));
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_exists (bool & exists)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_exists",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_exists,
                                 TR1::ref (exists) )); 
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_is_dir (bool & is_dir, int flags)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_is_dir",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_is_dir,
                                 TR1::ref (is_dir), flags )); 
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_is_entry (bool & is_entry, int flags)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_is_entry",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_is_entry,
                                 TR1::ref (is_entry), flags )); 
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_is_link (bool & is_link, int flags)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_is_link",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_is_link,
                                 TR1::ref (is_link), flags )); 
}

///////////////////////////////////////////////////////////////////////////////
template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_copy (saga::void_t & ret, 
    std::string source, int flags)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_copy",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_copy,
                                 TR1::ref (ret), source, flags)); 
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_link (saga::void_t & ret, 
    std::string target, int flags)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_link",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_link,
                                 TR1::ref(ret), target, flags)); 
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_move (saga::void_t & ret, 
    std::string source, int flags)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_move",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_move,
                                 TR1::ref (ret), source, flags)); 
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_remove (saga::void_t & ret, int flags)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_remove",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_remove,
                                 TR1::ref (ret), flags )); 
}

template <typename Base>
inline saga::task_type 
namespace_entry_cpi_impl<Base>::async_close (saga::void_t & ret, 
    double timeout)
{
    return (saga::adaptors::task("namespace_entry_cpi_impl<Base>::sync_close",
                                 this->shared_from_this(), 
                                 &namespace_entry_cpi_impl<Base>::sync_close,
                                 TR1::ref (ret), timeout )); 
}

/////////////////////////////////////////////////////////////////////////////
// bulk ops
#if ! defined (SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS)
// This adaptor implements the bulk functions for testing purposes only. Since 
// there is no principal need to do so, we allow these to be 'switched off'.
template <typename Base>
inline bool 
namespace_entry_cpi_impl<Base>::prepare_get_name(std::string& url, saga::uuid task_uuid)
{
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"Reached bulk_prepare"<<std::endl;
    }
    create_bulk_container_or_add_params(this, this->bulk_, 
        &namespace_entry_cpi_impl<Base>::bulk_get_name, TR1::ref(url), task_uuid);
    return true;
}

template <typename Base>
inline std::vector<saga::uuid> 
namespace_entry_cpi_impl<Base>::bulk_get_name(
                            std::vector<TR1::reference_wrapper<std::string> >& rets,
                            std::vector<saga::uuid>& uuids)
{
    
    SAGA_VERBOSE(SAGA_VERBOSE_LEVEL_BLURB) {
        std::cout<<"reached bulk_get_name for executing the bulk" << std::endl;
    }

      
    std::vector<TR1::reference_wrapper<std::string> >::iterator str_it = rets.begin();
    std::vector<saga::uuid>::iterator uuid_it = uuids.begin();
    
    std::vector<saga::uuid> r;
        
    while ( (str_it != rets.end()) && (uuid_it != uuids.end()))
    {
        TR1::reference_wrapper<std::string> ret = *str_it;
        this->sync_get_name(ret);
        r.push_back(*uuid_it);
        ++uuid_it;
        ++str_it;
    }   
    
    return r; 
}

#endif // ! SAGA_DEFAULT_FILE_ADAPTOR_NO_BULKS
#endif // ! SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS

#endif // !ADAPTORS_DEFAULT_FILE_DEFAULT_NAME_SPACE_ENTRY_IMPL_HPP

