//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <map>
#include <vector>
#include <ctime>

#include <boost/algorithm/string/predicate.hpp>

#include <saga/util.hpp>
#include <saga/exception.hpp>
#include <saga/url.hpp>
#include <saga/adaptors/task.hpp>
#include <saga/packages/job/adaptors/job.hpp>
#include <saga/packages/job/adaptors/job_self.hpp>
#include <saga/adaptors/attribute.hpp>

#include <impl/config.hpp>
#include <impl/engine/task_get_bulk_adaptor.hpp>

#include "default_job_service.hpp"

#ifndef  MAX_PATH
# define MAX_PATH _POSIX_PATH_MAX
#endif

///////////////////////////////////////////////////////////////////////////////
namespace {

    // wrapper for gethostname()
    inline std::string get_hostname()
    {
        char buffer[MAX_PATH] = { '\0' };
        gethostname(buffer, sizeof(buffer));
        return std::string(buffer);
    }

    // ensure to get a non-empty resource manager name
    inline std::string ensure_resourcemanager(std::string rm)
    {
        saga::url rm_url(rm);
        std::string host(rm_url.get_host());
        if (host.empty()) 
        {
            rm_url.set_scheme("fork");
            rm_url.set_host(get_hostname());
            rm = rm_url.get_url();
        }
        return rm;
    }
    
} // namespace

/*
 * constructor
 */
job_service_cpi_impl::job_service_cpi_impl (proxy                * p, 
                                            cpi_info       const & info,
                                            saga::ini::ini const & glob_ini, 
                                            saga::ini::ini const & adap_ini,
                                            TR1::shared_ptr<saga::adaptor> adaptor)
    : base_cpi (p, info, adaptor, cpi::Noflags)
{
    // check if we can handle this request
    instance_data data(this);
    if (!data->rm_.empty())
    {
        saga::url rm(data->rm_);
        std::string host(rm.get_host());
        if (!host.empty() && host != "localhost" && !boost::iequals(host, get_hostname()))
        {
            SAGA_ADAPTOR_THROW("Can't use remote hosts for job submission.",
                saga::NotImplemented);
        }

        std::string scheme(rm.get_scheme());
        if (!scheme.empty() && scheme != "fork" && scheme != "any")
        {
            SAGA_ADAPTOR_THROW("Can't use schemes others from 'any' or 'fork' " 
                "for job submission.", saga::NotImplemented);
        }
        
//         SAGA_VERBOSE (SAGA_VERBOSE_LEVEL_BLURB)
//         {
//           std::cout << "default job adaptor can create job_service" << std::endl; 
//         }
    }
}

/*
 * destructor
 */
job_service_cpi_impl::~job_service_cpi_impl (void)
{
}

/*
 * SAGA API functions
 */

///////////////////////////////////////////////////////////////////////////////
void 
job_service_cpi_impl::sync_create_job(saga::job::job& ret, saga::job::description jd)
{
    // make sure the executable path is given
    saga::attribute attr (jd);
    if (!attr.attribute_exists(saga::job::attributes::description_executable) ||
        attr.get_attribute(saga::job::attributes::description_executable).empty())
    {
        SAGA_ADAPTOR_THROW("Missing 'Executable' attribute on job description.",
            saga::BadParameter);
    }
    
    // get resource manager string
    std::string rm;
    { 
        instance_data data(this);
        rm = ensure_resourcemanager(data->rm_);
    }
    
    saga::job::job job = saga::adaptors::job(rm, jd, proxy_->get_session(), shared_from_this());

    // set the created attribute
    saga::adaptors::attribute jobattr (job);
    std::time_t current = 0;
    std::time(&current);
    jobattr.set_attribute(saga::job::attributes::created, ctime(&current));

    ret = job;
}

///////////////////////////////////////////////////////////////////////////////
namespace {

    // The command line splitting below is taken from the Boost.ProgramOptions
    // library.
    
    // Copyright Vladimir Prus 2002-2004.
    // Distributed under the Boost Software License, Version 1.0.
    // (See accompanying file LICENSE_1_0.txt
    // or copy at http://www.boost.org/LICENSE_1_0.txt)

    // Take a command line string and splits in into tokens, according
    // to the following rules
    //
    //  - Arguments are delimited by white space, which is either a space or 
    //    a tab.
    //  - A string surrounded by double quotation marks is interpreted as a 
    //    single argument, regardless of white space contained within. A 
    //    quoted string can be embedded in an argument. 
    //  - A double quotation mark preceded by a backslash, \", is interpreted 
    //    as a literal double quotation mark (").
    //  - Backslashes are interpreted literally, unless they immediately precede 
    //    a double quotation mark.
    //
    // These rules are a slight modification of the windows command line 
    // processor rules, as described here:
    //    http://article.gmane.org/gmane.comp.lib.boost.user/3005
    //    http://msdn.microsoft.com/library/en-us/vccelng/htm/progs_12.asp
    std::vector <std::string> split_commandline (std::string const & input)
    {
        using namespace std;
        vector <string> result;

        string::const_iterator i = input.begin ();
        string::const_iterator e = input.end   ();

        for ( /**/ ; i != e; ++i )
            if ( ! isspace ((unsigned char)*i) )
                break;
       
        if ( i != e ) 
        {
          string current;
          bool   inside_quoted   = false;
          int    backslash_count = 0;

          for (/**/; i != e; ++i) 
          {
            if (*i == '\\') 
            {
              // just count backslashes 
              ++backslash_count;
            }
            else if ( *i == '"' ) 
            {
              // '"' preceded by a backslash is a literal quote
              // the backslash which quoted is removed
              if ( backslash_count > 0 ) 
              {
                current += '"';
                --backslash_count;
              }
              // '"' not preceded by a backslash limits a quote
              else 
              {
                inside_quoted = ! inside_quoted;
              }
            } 
            else 
            {
              // Not quote or backslash. All accumulated backslashes should be
              // added
              if ( backslash_count ) 
              {
                current.append (backslash_count, '\\');
                backslash_count = 0;
              }

              if ( isspace ((unsigned char) *i) && ! inside_quoted ) 
              {
                // Space outside quoted section terminate the current argument
                result.push_back (current);
                current.resize   (0);
                
                for ( /**/ ; i != e && isspace ((unsigned char) *i); ++i ) 
                    /**/;

                --i;
              } 
              else 
              {                  
                current += *i;
              }
            }
          }

          // If we have trailing backslashes, add them
          if ( backslash_count )
          {
            current.append (backslash_count, '\\');
          }

          // If we have non-empty 'current' or we're still in quoted
          // section (even if 'current' is empty), add the last token.
          if ( ! current.empty () || inside_quoted )
          {
            result.push_back (current);        
          }
          // FIXME: we SHOULD thrown an exception here, non-matching quotes
          // are a BadParameter -- AM
        }
        return result;
    }

///////////////////////////////////////////////////////////////////////////////
}

void job_service_cpi_impl::sync_run_job(saga::job::job &    ret, 
                                        std::string    host, 
                                        std::string    commandline, 
                                        saga::job::ostream& in, 
                                        saga::job::istream& out, 
                                        saga::job::istream& err)
{
    std::vector<std::string> cmdline = split_commandline (commandline);
    std::string executable = cmdline[0];
    
    saga::job::description jd;
    jd.set_attribute(saga::job::attributes::description_candidatehosts, host);
    jd.set_attribute(saga::job::attributes::description_executable, executable);
    if (cmdline.size() > 1)
    {
        cmdline.erase(cmdline.begin());
        jd.set_vector_attribute(saga::job::attributes::description_arguments, 
                                cmdline);
    }
    jd.set_attribute(saga::job::attributes::description_interactive, 
        saga::attributes::common_true);

    // get resource manager string
    std::string rm;
    { 
        instance_data data(this);
        rm = ensure_resourcemanager(data->rm_);
    }
    
    // create a new job instance
    saga::job::job job = saga::adaptors::job(rm, jd, proxy_->get_session(), 
                                        shared_from_this());

    
    // set the created attribute 
    saga::adaptors::attribute attr (job);
    std::time_t current = 0;
    std::time(&current);
    attr.set_attribute(saga::job::attributes::created, ctime(&current));
    
    job.run();    // start this job
    /*
    std::cout << " ============ running job: " 
              << job.get_job_id () 
              << " - " 
              << job.get_state ()
              << "\n";
    */

#if defined(SAGA_USE_V2_TASK_MODEL)
    in = job.get_stdin();
    out = job.get_stdout();
    err = job.get_stderr();
#else
    job.get_stdin<saga::task_base::Sync>(in);
    job.get_stdout<saga::task_base::Sync>(out);
    job.get_stderr<saga::task_base::Sync>(err);
#endif

    ret = job;
}

///////////////////////////////////////////////////////////////////////////////
//  return the list of jobs created by this job_service
void job_service_cpi_impl::sync_list(std::vector<std::string>& list_of_jobids)
{
    adaptor_data_type data(this);
    list_of_jobids = data->list_jobs();
}

void job_service_cpi_impl::sync_get_job(saga::job::job & ret, std::string jobid)
{
    // get resource manager string
    std::string rm;
    { 
        instance_data data(this);
        rm = ensure_resourcemanager(data->rm_);
    }
    saga::job::job job = saga::adaptors::job(rm, jobid, proxy_->get_session(), 
                                        shared_from_this());

    // set initial attribute if this is a new job
    if (saga::job::New == job.get_state()) {
        saga::adaptors::attribute attr (job);
        std::time_t current = 0;
        std::time(&current);
        attr.set_attribute(saga::job::attributes::created, ctime(&current));
    }
    
    ret = job;
}

///////////////////////////////////////////////////////////////////////////////
namespace 
{
    // keep track of the time this adaptor was loaded
    struct started_time
    {
        started_time()
        :   started_(0)
        {
            std::time(&started_);
        }
        
        std::time_t started_;
    };
    static started_time started;
}
 
void job_service_cpi_impl::sync_get_self(saga::job::self& ret)
{
    // get resource manager string
    std::string rm;
    { 
        instance_data data(this);
        rm = ensure_resourcemanager(data->rm_);
    }
    
    // get at our jobid
    SAGA_OSSTREAM strm;
    strm << "[" << rm << "]-[" << get_hostname() << ":" << getpid() << "]";
    std::string jobid = SAGA_OSSTREAM_GETSTRING(strm);
    
    // create the job_self instance
    saga::job::self job = saga::adaptors::self(rm, jobid, proxy_->get_session(), 
        shared_from_this());

    // set/modify some of the job attributes
    saga::adaptors::attribute jobattr (job);
    std::time_t current = started.started_;
    jobattr.set_attribute(saga::job::attributes::created, ctime(&current));

    std::vector<std::string> hosts;
    hosts.push_back(get_hostname());
    jobattr.set_vector_attribute(saga::job::attributes::executionhosts, hosts);

    ret = job;
}

///////////////////////////////////////////////////////////////////////////////
#if !defined (SAGA_DEFAULT_JOB_ADAPTOR_NO_ASYNCS)
// This adaptor implements the async functions based on its own synchronous
// functions for testing purposes only. Since there is no principal need
// to do so, we allow these to be 'switched off'.

saga::task_type 
job_service_cpi_impl::async_create_job(saga::job::job& ret, saga::job::description jd)
{
    return saga::adaptors::task("job_service_cpi_impl::async_create_job",
        shared_from_this(), &job_service_cpi_impl::sync_create_job, 
        TR1::ref(ret), jd);
}

saga::task_type 
job_service_cpi_impl::async_run_job(saga::job::job& ret, std::string host, 
    std::string commandline, saga::job::ostream& in, saga::job::istream& out, 
    saga::job::istream& err)
{
    return saga::adaptors::task("job_service_cpi_impl::async_run_job",
        shared_from_this(), &job_service_cpi_impl::sync_run_job, 
        TR1::ref(ret), host, commandline, TR1::ref(in), TR1::ref(out),
        TR1::ref(err));
}

saga::task_type 
job_service_cpi_impl::async_list(std::vector<std::string>& list_of_jobids)
{
    return saga::adaptors::task("job_service_cpi_impl::async_list",
        shared_from_this(), &job_service_cpi_impl::sync_list, 
        TR1::ref(list_of_jobids));
}

saga::task_type 
job_service_cpi_impl::async_get_job(saga::job::job& ret, std::string jobid)
{
    return saga::adaptors::task("job_service_cpi_impl::async_get_job",
        shared_from_this(), &job_service_cpi_impl::sync_get_job, 
        TR1::ref(ret), jobid);
}

saga::task_type 
job_service_cpi_impl::async_get_self(saga::job::self& ret)
{
    return saga::adaptors::task("job_service_cpi_impl::async_get_self",
        shared_from_this(), &job_service_cpi_impl::sync_get_self, 
        TR1::ref(ret));
}

#endif  // !defined (SAGA_DEFAULT_JOB_ADAPTOR_NO_ASYNCS)
