//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(ADAPTORS_DEFAULT_LOGICAL_FILE_COMMON_HELPERS_HPP)
#define ADAPTORS_DEFAULT_LOGICAL_FILE_COMMON_HELPERS_HPP

#include <list>
#include <string>

#include <boost/config.hpp>
#include <boost/algorithm/string/predicate.hpp>
#if defined(BOOST_HAS_UNISTD_H)
#include <unistd.h>
#endif
#if defined(BOOST_WINDOWS)
#include <windows.h>
#endif

#include <saga/adaptors/utils/ini/ini.hpp>
#include <saga/context.hpp>
#include <saga/session.hpp>
#include <saga/url.hpp>

#include "logicalfile_database.hpp"
#include "connectstring.hpp"

#ifndef  MAX_PATH
# define MAX_PATH _POSIX_PATH_MAX
#endif

///////////////////////////////////////////////////////////////////////////////
namespace logicalfile { namespace detail
{
    ///////////////////////////////////////////////////////////////////////////
    // wrapper for gethostname()
    inline std::string 
    get_hostname()
    {
        char buffer[MAX_PATH] = { '\0' };
        gethostname(buffer, sizeof(buffer));
        return std::string(buffer);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //  retrieve from ini file the type of database to use
    inline database::db_type 
    get_db_type(saga::ini::ini const& ini)
    {
    // try to find database back end
        if (ini.has_section("preferences")) {
        saga::ini::ini prefs = ini.get_section ("preferences");
        std::string dbtype(prefs.has_entry("dbtype") ? prefs.get_entry("dbtype") : "");
            
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_SQLITE3)
            if (dbtype == "sqlite3")
                return database::db_sqlite;
#endif
        // handle postgresql
            if (dbtype == "postgresql")
#if defined(SAGA_DEFAULT_ADVERT_ADAPTOR_USE_POSTGRESQL)
                return database::db_postgresql;
#else
            {
                SAGA_OSSTREAM strm;
                strm << "advert::advert_cpi_impl::init: "
                        "PostgreSQL backend has not been configured.";
                SAGA_ADAPTOR_THROW_NO_CONTEXT(SAGA_OSSTREAM_GETSTRING(strm), 
                    saga::BadParameter);
            }
#endif

      // handle mysql
            if (dbtype == "mysql")
#if defined(SAGA_DEFAULT_ADVERT_ADAPTOR_USE_MYSQL)
                return database::db_mysql;
#else
            {
                SAGA_OSSTREAM strm;
                strm << "advert::advert_cpi_impl::init: "
                        "MySQL backend has not been configured.";
                SAGA_ADAPTOR_THROW_NO_CONTEXT(SAGA_OSSTREAM_GETSTRING(strm), 
                    saga::BadParameter);
            }
#endif

#if !defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_SQLITE3)
            SAGA_OSSTREAM strm;
            strm << "logicalfile::logical_directory_cpi_impl::init: "
                    "unknown database backend specified in configuration: "
                 << dbtype;
            SAGA_ADAPTOR_THROW_NO_CONTEXT(SAGA_OSSTREAM_GETSTRING(strm), 
                saga::BadParameter);
#endif
        }

#if !defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_SQLITE3)
    // throw error 
        SAGA_OSSTREAM strm;
        strm << "logicalfile::logical_directory_cpi_impl::init: "
                "no database back end specified in configuration.";
        SAGA_ADAPTOR_THROW_NO_CONTEXT(SAGA_OSSTREAM_GETSTRING(strm), 
            saga::BadParameter);
#endif
        return database::db_sqlite;   // default to sqlite, if possible
    }

#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_POSTGRESQL)
    // generate connectstring for a postgresql back end
    template <typename InstanceData>
    inline std::string
    construct_pg_connect(std::string const& connect, saga::session const& s,
        saga::impl::proxy* p)
    {
        InstanceData data(p);
        connectstring connectstr(connect);
        
    // set username and password from a (possibly) existing context
        std::list<saga::context> ctxs(s.list_contexts());
        std::list<saga::context>::iterator end = ctxs.end();
        for (std::list<saga::context>::iterator it = ctxs.begin(); 
             it != end; ++it)
        {
            if (saga::context::UserPass == (*it).get_ctype()) {
                connectstr.set_value("user",
                    (*it).get_attribute(saga::attributes::context_user_name));
                connectstr.set_value("password",
                    (*it).get_attribute(saga::attributes::context_user_pass));
                break;
            }
        }

    // set the host name and port number
        saga::url lfn_url(data->location_);
        std::string host(lfn_url.get_host());
        if (!host.empty())
            connectstr.set_value("host", host);
        
        int port = lfn_url.get_port();
        if (-1 != port) 
            connectstr.set_value("port", boost::lexical_cast<std::string>(port));
        
    // set username and password from the url
        std::string username(lfn_url.get_username());
        if (!username.empty())
            connectstr.set_value("user", username);

        std::string password(lfn_url.get_password());
        if (!password.empty())
            connectstr.set_value("password", password);

    // make sure a dbname is given
        if (!connectstr.has_value("dbname"))
            connectstr.set_value("dbname", "logicalfiledb");
            
        return connectstr.get();
    }
#endif // SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_POSTGRESQL

    // compose the connect string to the backend
    template <typename InstanceData>
    inline std::string 
    get_db_connectstring(saga::ini::ini const& ini, saga::session const& s,
        saga::impl::proxy* p)
    {
    // try to find database back end
        if (ini.has_section("preferences")) {
        saga::ini::ini prefs = ini.get_section ("preferences");
        std::string dbconnect(prefs.has_entry("dbconnect") ? 
            prefs.get_entry("dbconnect") : "");
            
            if (!dbconnect.empty()) {
                switch (get_db_type(ini)) {
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_SQLITE3)
                case database::db_sqlite:
                    return dbconnect;
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_POSTGRESQL)
                case database::db_postgresql:
                    return construct_pg_connect<InstanceData>(dbconnect, s, p);
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_MYSQL)
                case database::db_mysql:
                    return construct_mysql_connect<InstanceData>(dbconnect, s, p);
#endif
                default:
                    return dbconnect;
                }
            }                
        }

        switch (get_db_type(ini)) {
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_SQLITE3)
        case database::db_sqlite:
            return "logicalfile.db";
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_POSTGRESQL)
        case database::db_postgresql:
            return "dbname=logicalfiledb";
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_MYSQL)
        case database::db_mysql:
            return "dbname=logicalfiledb";
#endif
            
        default:
            break;
        }

    // throw error 
        SAGA_OSSTREAM strm;
        strm << "logicalfile::logical_directory_cpi_impl::init: "
                "no database back end specified in configuration.";
        SAGA_ADAPTOR_THROW_NO_CONTEXT(SAGA_OSSTREAM_GETSTRING(strm), 
            saga::BadParameter);
        return "";    // unreachable code
    }
    
    // generate a regex matching sub-clause for different db back ends
    inline std::string
    get_match (database::db_type type, char const* lhs, char const* rhs)
    {
        switch(type) {
        // these databases use a special user define function 'match()'
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_SQLITE3)
        case database::db_sqlite:
            return std::string("match(") + lhs + ", " + rhs + ")";
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_POSTGRESQL)
        case database::db_postgresql:
            return std::string("match(") + lhs + ", " + rhs + ")";
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_MYSQL)
        // MySql natively supports regular expressions
        case database::db_mysql:
            return std::string(lhs) + " REGEXPR " + rhs;
#endif

        default:
            break;
        }
        
    // throw error 
        SAGA_OSSTREAM strm;
        strm << "logicalfile::detail::get_match: "
                "no database back end specified in configuration.";
        SAGA_ADAPTOR_THROW_NO_CONTEXT(SAGA_OSSTREAM_GETSTRING(strm), 
            saga::BadParameter);
        return "";    // unreachable code
    }

    ///////////////////////////////////////////////////////////////////////////
    inline std::string 
    get_filepath(std::string const& loc)
    {
        saga::url lfn_url(loc);
        return lfn_url.get_path();
    }

    ///////////////////////////////////////////////////////////////////////////
    // verify the hostname will be understood by our adaptor
    inline bool 
    host_is_valid(database::db_type type, std::string const& host)
    {
        switch (type) {
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_SQLITE3)
        case database::db_sqlite:
            return host == "localhost" || boost::iequals(host, detail::get_hostname());
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_POSTGRESQL)
        case database::db_postgresql:
            return true;
#endif
#if defined(SAGA_DEFAULT_LOGICALFILE_ADAPTOR_USE_MYSQL)
        case database::db_mysql:
            return true;
#endif

        default:
            break;
        }
        return false;
    }
    
    // convert a string attribute in a vector attribute by splitting the 
    // attribute value at ',' symbols
    inline void
    tokenize_values(std::string val, std::vector<std::string>& values)
    {
        typedef boost::char_separator<char> char_separator_type;
        typedef boost::tokenizer<char_separator_type> tokenizer_type;
        
        char_separator_type sep (",");
        tokenizer_type tok (val, sep);
        tokenizer_type::iterator tok_end = tok.end();
        for (tokenizer_type::iterator tit = tok.begin (); tit != tok_end; ++tit)
        {
            values.push_back (*tit);
        }
    }

    // convert a vector attribute into a single value by concatenating the 
    // values into a list separated by ',' symbols
    inline std::string 
    concat_values(std::vector<std::string> const& values)
    {
        std::string result;
        std::vector<std::string>::const_iterator end = values.end();
        for (std::vector<std::string>::const_iterator it = values.begin();
             it != end; /**/)
        {
            result += *it;
            if (++it != end)
                result += ",";
        }
        return result;
    }

    // convert a SAGA pattern into a TR1::regex 
    inline std::string 
    regex_from_character_set(std::string::const_iterator& it, 
        std::string::const_iterator end)
    {
        std::string::const_iterator start = it;
        std::string result(1, *it);  // copy '['
        if (*++it == '!') {
            result.append(1, '^');   // negated character set
        }
        else if (*it == ']') {
            SAGA_ADAPTOR_THROW_NO_CONTEXT(
                "Invalid pattern (empty character set) at: " + 
                std::string(start, end), saga::NoSuccess);
        }
        else {
            result.append(1, *it);   // append this character
        }
        
        // copy while in character set
        while (++it != end) {
            result.append(1, *it);
            if (*it == ']')
                break;
        }
        
        if (it == end || *it != ']') {
            SAGA_ADAPTOR_THROW_NO_CONTEXT(
                "Invalid pattern (missing closing ']') at: " + 
                std::string(start, end), saga::NoSuccess);
        }

        return result;
    }
    
    inline std::string 
    regex_from_string_choice(std::string::const_iterator& it, 
        std::string::const_iterator end)
    {
        std::string::const_iterator start = it;
        std::string result(1, '(');

        // copy while in string choice
        bool has_one = false;
        while (++it != end && *it != '}') {
            if (*it == ',') {
                if (!has_one) {
                    SAGA_ADAPTOR_THROW_NO_CONTEXT(
                        "Invalid pattern (empty string alternative) at: " + 
                        std::string(start, end), saga::NoSuccess);
                }
                result.append(1, '|');
                has_one = false;
            }
            else {
                result.append(1, *it);
                has_one = true;
            }
        }
        
        if (it == end || *it != '}') {
            SAGA_ADAPTOR_THROW_NO_CONTEXT(
                "Invalid pattern (missing closing '}') at: " + 
                std::string(start, end), saga::NoSuccess);
        }
        
        result.append(1, ')');
        return result;
    }
    
    inline std::string 
    regex_from_pattern(std::string const& pattern)
    {
        std::string result;
        std::string::const_iterator end = pattern.end();
        for (std::string::const_iterator it = pattern.begin(); it != end; ++it)
        {
            char c = *it;
            switch (c) {
            case '*':
                result.append(".*");
                break;
                
            case '?':
                result.append(1, '.');
                break;
                
            case '[':
                result.append(regex_from_character_set(it, end));
                break;
                
            case '{':
                result.append(regex_from_string_choice(it, end));
                break;
                
            case '\\':
                if (++it == end) {
                    SAGA_ADAPTOR_THROW_NO_CONTEXT(
                        "Invalid escape sequence at: " + pattern,
                        saga::NoSuccess);
                }
                result.append(1, *it);
                break;
                
            default:
                result.append(1, c);
                break;
            }
        }
        return result;
    }
    
    inline void
    regex_from_patterns(std::vector<std::string> const& patterns,
        std::vector<std::string>& result)
    {
        std::transform(patterns.begin(), patterns.end(), std::back_inserter(result),
            regex_from_pattern);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    inline void 
    parse_pattern(saga::impl::v1_0::cpi const* target, 
        std::string const& pattern, std::string& kpat, std::string& vpat)
    {
        std::string::size_type p = pattern.find_first_of("=");
        if (std::string::npos == p) {
            SAGA_ADAPTOR_THROW_VERBATIM(target, 
                "Pattern has invalid format: " + pattern, saga::BadParameter);
        }
        kpat = pattern.substr(0, p);
        vpat = pattern.substr(p+1);
    }
    
    template <typename Iterator>
    inline void 
    parse_patterns(saga::impl::v1_0::cpi const* target, 
        std::vector<std::string> const& patterns, Iterator kit, Iterator vit)
    {
        std::vector<std::string>::const_iterator end = patterns.end();
        for (std::vector<std::string>::const_iterator it = patterns.begin(); 
             it != end; ++it, ++kit, ++vit)
        {
            std::string kpat, vpat;
            parse_pattern(target, *it, kpat, vpat);
            *kit = kpat;
            *vit = vpat;
        }
    }
    
    inline void 
    parse_patterns(saga::impl::v1_0::cpi const* target, 
        std::vector<std::string> const& patterns, std::vector<std::string>& kpat, 
        std::vector<std::string>& vpat)
    {
        parse_patterns(target, patterns, std::back_inserter(kpat),
            std::back_inserter(vpat));
    }
    
///////////////////////////////////////////////////////////////////////////////
}}  // namespace logicalfile::detail

#endif // !ADAPTORS_DEFAULT_LOGICAL_FILE_COMMON_HELPERS_HPP
