//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 
#include <map>
#include <vector>

#include <boost/assert.hpp>
#include <boost/tokenizer.hpp>

#include <saga/util.hpp>
#include <saga/exception.hpp>
#include <saga/url.hpp>
#include <saga/adaptors/task.hpp>
#include <saga/file.hpp>

#include <impl/config.hpp>
#include <impl/engine/task_get_bulk_adaptor.hpp>
#include <impl/exception_list.hpp>

#include "config.hpp"
#include "logicalfile.hpp"
#include "common_helpers.hpp"

///////////////////////////////////////////////////////////////////////////////
// wrap database calls and catch all exceptions
#define SAGA_LOGICALFILE_DBACTION(name, f)                                    \
    try {                                                                     \
        (f);                                                                  \
    }                                                                         \
    catch (saga::adaptors::exception const& e) {                              \
    /* just rethrow using this cpi instance */                                \
        saga::impl::exception_list exceptions;                                \
        exceptions.add(e);                                                    \
        SAGA_ADAPTOR_THROW(exceptions.get_message(), exceptions.get_error()); \
    }                                                                         \
    catch (std::exception const& e) {                                         \
    /* wrap this error in a SAGA adaptor exception */                         \
        SAGA_OSSTREAM strm;                                                   \
        strm << "logicalfile::logical_file_cpi_impl::" #name ": "             \
             << e.what();                                                     \
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::NoSuccess);   \
    }                                                                         \
    /**/
    
///////////////////////////////////////////////////////////////////////////////
namespace logicalfile
{

///////////////////////////////////////////////////////////////////////////////
//  constructor
logical_file_cpi_impl::logical_file_cpi_impl (proxy* p, cpi_info const& info,
        saga::ini::ini const& glob_ini, saga::ini::ini const& adap_ini,
        TR1::shared_ptr<saga::adaptor> adaptor)
:   base_cpi (p, info, adaptor, cpi::Noflags),
    db_(detail::get_db_type(adap_ini), 
        detail::get_db_connectstring<instance_data>(adap_ini, p->get_session(), p)),
    thrd_(NULL), 
    cond_(NULL),
    thread_alive_(false), cancel_thread_(false)
{
    saga::url lfn_url;
    std::string url;
    {
        // first usage of this adaptor
        instance_data data (this);
        lfn_url = data->location_;
        url = lfn_url.get_path();

        // we understand lfn's only for localhost
        std::string host(lfn_url.get_host());
        if (!host.empty() && !detail::host_is_valid(db_.get_dbtype(), host))
        {
            SAGA_OSSTREAM strm;
            strm << "logicalfile::logical_file_cpi_impl::init: "
                    "cannot handle logical file name: " 
                 << lfn_url.get_url();
            SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::IncorrectURL);
        }

        // we understand only 'any' and 'lfn' url schemes
        std::string scheme(lfn_url.get_scheme());
        if (!scheme.empty() && scheme != "lfn" && scheme != "any")
        {
            SAGA_OSSTREAM strm;
            strm << "logicalfile::logical_file_cpi_impl::init: "
                    "cannot handle logical file name: " << lfn_url.get_url();
            SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::IncorrectURL);
        }
        
        saga::logical_file::flags mode = (saga::logical_file::flags)data->mode_;
        if (((mode & saga::logical_file::Create) || 
             (mode & saga::logical_file::CreateParents)) && 
            (mode & saga::logical_file::Exclusive)) 
        {
            if (db_.exists_entry(url))
            {
                SAGA_OSSTREAM strm;
                strm << "logicalfile::logical_file_cpi_impl::init: "
                        "replica entry already exists: " << lfn_url.get_url();
                SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::AlreadyExists);
            }
        }

        // Create replica entry if saga::logical_file::create flag is given
        if ((mode & saga::logical_file::Create) || 
            (mode & saga::logical_file::CreateParents))
        {
            SAGA_LOGICALFILE_DBACTION(logical_file_cpi_impl, 
                (db_.create_entry(url, mode & saga::logical_file::CreateParents)));
            data->mode_ &= ~saga::logical_file::Create;   // no need to create this entry twice
        }
    }
    
    bool result = false;

    SAGA_LOGICALFILE_DBACTION(logical_file_cpi_impl, 
        (result = db_.exists_entry(url)));
        
    if (!result)
    { 
        SAGA_OSSTREAM strm;
        strm << "logicalfile::logical_file_cpi_impl::init: "
                "replica entry does not exist: " << lfn_url.get_url();
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::BadParameter);
    }
    
    // set up thread monitoring this entry in the database
    setup_monitoring_thread();
}

///////////////////////////////////////////////////////////////////////////////
void logical_file_cpi_impl::setup_monitoring_thread()
{
    mutex_type::scoped_lock lock(mtx_);
    
    BOOST_ASSERT(NULL == cond_ && NULL == thrd_); // should be called only once

    // set up separate thread watching for job termination
    thread_alive_ = true;
    cancel_thread_ = false;
    try {
        cond_ = new boost::condition();
        thrd_ = new boost::thread(TR1::bind(&logical_file_cpi_impl::monitor_entry, this));
        cond_->wait(lock);
    }
    catch (...) {
        thread_alive_ = false;
        cancel_thread_ = true;
        delete thrd_;
        thrd_ = NULL;
    }
    delete cond_;
    cond_ = NULL;
}

///////////////////////////////////////////////////////////////////////////////
namespace 
{
    struct handle_flag
    {
        handle_flag(logical_file_cpi_impl::mutex_type& mtx, bool& flag) 
          : mtx_(mtx), flag_(flag)
        {
        }
        ~handle_flag()
        {
            logical_file_cpi_impl::mutex_type::scoped_lock l(mtx_);
            flag_ = false;
        }
        
        logical_file_cpi_impl::mutex_type& mtx_;
        bool& flag_;
    };
}

void logical_file_cpi_impl::monitor_entry()
{
    handle_flag handler(mtx_, thread_alive_);   // 'thread_alive_ = false' on exit
    
    bool first = true;
    try {
        std::string key("utime");
        std::string adname;
        {
            instance_data data (this);
            adname = detail::get_filepath(data->location_);
        }
        
        std::string value;
        bool exists = false;
            
    // fetch modification time
        std::string utime;
        do {
            mutex_type::scoped_lock l(mtx_);

        // read entry
            exists = false;
            SAGA_LOGICALFILE_DBACTION(attribute_exists, 
                (db_.attribute_exists(adname, key, exists)));
            if (!exists) 
                break;    // entry has been deleted
                
            bool is_vector = false;
            SAGA_LOGICALFILE_DBACTION(get_attribute, 
                (db_.get_attribute(adname, key, utime, is_vector)));

        // try to find out, if entry has been changed
            if (first) {
                value = utime;
                first = false;
                if (NULL != cond_) 
                    cond_->notify_one();        // initialization is complete
            }
            else {
                boost::xtime xt;
                boost::xtime_get(&xt, boost::TIME_UTC);
                xt.nsec += 10000000;            // sleep max. 10ms
                if (cancel_thread_ || cancel_thrd_.timed_wait(l, xt))
                    return;                     // thread has been canceled

            // if the modification time has been changed, fire metric
                if (utime != value) {
                    saga::monitorable monitor (this->proxy_);
                    saga::adaptors::metric m (
                        monitor.get_metric(saga::metrics::logical_file_modified));
                    m.set_attribute(saga::attributes::metric_value, adname);
                    m.fire();
                    value = utime;
                }
            }
        } while (true);
            
    // if entry doesn't exist anymore, fire deleted metric
        BOOST_ASSERT(!exists);
        saga::monitorable monitor (this->proxy_);
        saga::adaptors::metric m (
            monitor.get_metric(saga::metrics::logical_file_deleted));
        m.set_attribute(saga::attributes::metric_value, adname);
        m.fire();
    }
    catch (...) {
        // just ignore any error in the monitoring thread
        if (first) {
            mutex_type::scoped_lock l(mtx_);
            if (NULL != cond_) 
                cond_->notify_one();        // initialization is complete
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
//  destructor
logical_file_cpi_impl::~logical_file_cpi_impl (void)
{
    mutex_type::scoped_lock lock(mtx_);

    if (NULL != thrd_) {
        if (thread_alive_) {
            cancel_thread_ = true;        // break thread
            cancel_thrd_.notify_one();
            lock.unlock();
            thrd_->join();
        }
        delete thrd_;
        thrd_ = NULL;
    }

    SAGA_LOGICALFILE_DBACTION(~logical_file_cpi_impl, db_.close(false));
}

///////////////////////////////////////////////////////////////////////////////
//  SAGA CPI functions 

///////////////////////////////////////////////////////////////////////////////
// attribute functions
void 
logical_file_cpi_impl::sync_attribute_exists(bool& ret, std::string key)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(attribute_exists, 
        (db_.attribute_exists(lfn, key, ret)));
}

void 
logical_file_cpi_impl::sync_attribute_is_readonly(bool& ret, std::string key)
{
    instance_data data (this);
    ret = !(data->mode_ & saga::logical_file::Read);
}

void 
logical_file_cpi_impl::sync_attribute_is_writable(bool& ret, std::string key)
{
    instance_data data (this);
    ret = data->mode_ & saga::logical_file::Write;
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::check_permissions(saga::logical_file::flags flags,
    char const* name, std::string const& lfn)
{
    instance_data data (this);
    if (!(data->mode_ & flags)) {
        SAGA_OSSTREAM strm;
        strm << "logicalfile::logical_file_cpi_impl::" << name << ": "
                "could not access(" 
             << ((flags == saga::logical_file::Read) ? "read" : "write")
             << ") attribute for replica entry: " 
             << lfn;
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), 
            saga::PermissionDenied);
    }
}

void 
logical_file_cpi_impl::sync_attribute_is_vector(bool& ret, std::string key)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }
    
    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Read, "attribute_is_vector", lfn);
    
    // read attribute from db 
    mutex_type::scoped_lock lock(mtx_);
    std::string val;
    SAGA_LOGICALFILE_DBACTION(attribute_is_vector, 
        (db_.get_attribute(lfn, key, val, ret)));
}

void 
logical_file_cpi_impl::sync_attribute_is_extended(bool& ret, std::string key)
{
    instance_data data (this);
    ret = data->mode_ & saga::logical_file::Write;
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::sync_get_attribute(std::string& ret, std::string key)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }
    
    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Read, "get_attribute", lfn);

    // read attribute from db 
    mutex_type::scoped_lock lock(mtx_);
    std::string val;
    bool is_vector = false;

    SAGA_LOGICALFILE_DBACTION(get_attribute, 
        (db_.get_attribute(lfn, key, val, is_vector)));

    if (!is_vector) {
        ret = val;
    }
    else {
        SAGA_OSSTREAM strm;
        strm << "logicalfile::logical_file_cpi_impl::sync_get_attribute: "
                "attribute '" << key << "' is a vector attribute for: " << lfn;
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), 
            saga::IncorrectState);
    }
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::sync_get_vector_attribute(std::vector<std::string>& ret, 
    std::string key)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }
    
    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Read, "get_vector_attribute", lfn);

    // read attribute from db 
    std::string val;
    bool is_vector = false;
    bool result    = false;

    {
        mutex_type::scoped_lock lock(mtx_);
        SAGA_LOGICALFILE_DBACTION(get_vector_attribute, 
            (result = db_.get_attribute(lfn, key, val, is_vector)));
    }
    if (result) {
        if (is_vector) {
            detail::tokenize_values(val, ret);
        }
        else {
            SAGA_OSSTREAM strm;
            strm << "logicalfile::logical_file_cpi_impl::sync_get_vector_attribute: "
                    "attribute '" << key << "' is not a vector attribute for: " 
                 << lfn;
            SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), 
                saga::IncorrectState);
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::sync_set_attribute(saga::void_t&, std::string key, 
    std::string val)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Write, "set_attribute", lfn);
    
    // write attribute to db
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(set_attribute, 
        (db_.set_attribute(lfn, key, val, false)));
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::sync_set_vector_attribute(saga::void_t&, std::string key, 
    std::vector<std::string> val)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Write, "get_vector_attribute", lfn);
    
    // write attribute to db
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(set_vector_attribute, 
        (db_.set_attribute(lfn, key, detail::concat_values(val), true)));
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::sync_remove_attribute(saga::void_t&, std::string key)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Write, "remove_attribute", lfn);
    
    // remove attribute from db
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(remove_attribute, 
        (db_.remove_attribute(lfn, key)));
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::sync_list_attributes(std::vector<std::string>& keys)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Read, "list_attributes", lfn);
    
    // get attributes from db
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(list_attributes, 
        (db_.list_attributes(lfn, keys)));
}

///////////////////////////////////////////////////////////////////////////////
void 
logical_file_cpi_impl::sync_find_attributes(std::vector<std::string>& keys, 
    std::string pattern)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Read, "find_attributes", lfn);
    
    // extract patterns from 'key=value' expression
    std::string kpat, vpat;
    detail::parse_pattern(this, pattern, kpat, vpat);
    
    // look for matching keys in the database
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(find_attributes, 
        (db_.find_attributes(lfn, detail::regex_from_pattern(kpat), 
            detail::regex_from_pattern(vpat), &keys)));
}

///////////////////////////////////////////////////////////////////////////////
// namespace_entry functions
void 
logical_file_cpi_impl::sync_get_url(std::string& url)
{
    {
        mutex_type::scoped_lock lock(mtx_);
        SAGA_LOGICALFILE_DBACTION(get_url, (db_.verify_status()));  // throws when db has been closed
    }    
    instance_data data (this);
    url = data->location_;
}

// spec: "For not-directory type entries, the URL returned on get_url MUST
//        equal the concatenation of the return values of get_cwd() and 
//        get_name()"
void 
logical_file_cpi_impl::sync_get_cwd(std::string& url)
{
    {
        mutex_type::scoped_lock lock(mtx_);
        SAGA_LOGICALFILE_DBACTION(get_cwd, (db_.verify_status()));  // throws when db has been closed
    }
    saga::url lfn;
    {
        instance_data data (this);
        lfn = data->location_;
    }
    
    std::string path(lfn.get_path());
    std::string::size_type p = path.find_last_of('/');
    if (p != std::string::npos)
        lfn.set_path(path.substr(0, p+1));
    
    url = lfn.get_url();
}

void 
logical_file_cpi_impl::sync_get_name(std::string& url)
{
    {
        mutex_type::scoped_lock lock(mtx_);
        SAGA_LOGICALFILE_DBACTION(get_name, (db_.verify_status()));  // throws when db has been closed
    }
    
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    std::string::size_type p = lfn.find_last_of('/');
    if (p != std::string::npos)
        url = lfn.substr(p);
    else
        url.clear();
}

void 
logical_file_cpi_impl::sync_is_dir(bool& ret, int flags)
{
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(is_dir, (db_.verify_status()));     // throws when db has been closed
    ret = false;
}

void 
logical_file_cpi_impl::sync_is_entry(bool& ret, int flags)
{
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(is_entry, (db_.verify_status()));   // throws when db has been closed
    ret = true;
}

void 
logical_file_cpi_impl::sync_is_link(bool& ret, int flags)
{
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(is_link, (db_.verify_status()));    // throws when db has been closed
    ret = false;                // we don't support links
}

void 
logical_file_cpi_impl::sync_remove(saga::void_t&, int flags)
{
    if (flags & saga::logical_file::Recursive) {
        SAGA_ADAPTOR_THROW(
            "logicalfile::logical_file_cpi_impl::sync_remove: "
            "Recursive flag was specified while attempting to delete a file",
            saga::BadParameter);
    }

    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Write, "remove", lfn);
    
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(remove, (db_.remove(lfn), db_.close()));
}

void logical_file_cpi_impl::sync_close(saga::void_t&, double)
{
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(close, db_.close());
}


///////////////////////////////////////////////////////////////////////////////
// logical_file functions
void 
logical_file_cpi_impl::sync_list_locations(std::vector<std::string>& locations)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Read, "list_locations", lfn);
    
    // get locations from db    
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(list_locations, (db_.get_entries(lfn, locations)));
}

void 
logical_file_cpi_impl::sync_add_location(saga::void_t&, std::string location)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Write, "add_location", lfn);
    
    // add new replica
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(add_location, (db_.add_entry(lfn, location)));
}

void 
logical_file_cpi_impl::sync_remove_location(saga::void_t&, std::string location)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Write, "remove_location", lfn);
    
    // remove replica
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(remove_location, (db_.remove_entry(lfn, location)));
}

void 
logical_file_cpi_impl::sync_update_location(saga::void_t&, 
    std::string oldlocation, std::string newlocation)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Write, "update_location", lfn);
    
    // remove old and add new replica
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(update_location, 
        (db_.update_entry(lfn, oldlocation, newlocation)));
}

void 
logical_file_cpi_impl::sync_replicate(saga::void_t&, std::string location, 
    int mode)
{
    std::string lfn;
    {
        instance_data data (this);
        lfn = detail::get_filepath(data->location_);
    }

    // check permissions (throws on mismatch)
    check_permissions(saga::logical_file::Read, "replicate", lfn);
    check_permissions(saga::logical_file::Write, "replicate", lfn);

    // get replicas for the given replica entry
    std::vector<std::string> locations;
    
    bool result = false;
    {
        mutex_type::scoped_lock lock(mtx_);
        SAGA_LOGICALFILE_DBACTION(replicate, 
            (result = db_.get_entries(lfn, locations)));
    }    
    if (!result) {
        SAGA_OSSTREAM strm;
        strm << "logicalfile::logical_file_cpi_impl::replicate: "
                "replica entry: '" << lfn << "' does not contain any replicas.";
        SAGA_ADAPTOR_THROW(SAGA_OSSTREAM_GETSTRING(strm), saga::IncorrectState);
    }
    BOOST_ASSERT(!locations.empty());

    // loop over all replicas trying to copy (replicate) the file
    saga::impl::exception_list exceptions;
    bool copied_file = false;

    std::vector<std::string>::iterator end = locations.end();
    for (std::vector<std::string>::iterator it = locations.begin(); 
         it != end && !copied_file; ++it)
    {
        try {
            saga::filesystem::file f(*it);
            f.copy(location, mode);
        }
        catch (saga::exception const& e) {
            exceptions.add(e);
            continue;
        }
        copied_file = true;
    }
    
    if (!copied_file) {
        // rethrow the whole error stack 
        SAGA_ADAPTOR_THROW(exceptions.get_message(), exceptions.get_error());
    }
    
    mutex_type::scoped_lock lock(mtx_);
    SAGA_LOGICALFILE_DBACTION(replicate, (db_.add_entry(lfn, location)));
}

///////////////////////////////////////////////////////////////////////////////
}   // namespace logicalfile

