//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef ADAPTORS_DEFAULT_LOGICAL_FILE_LOGICALFILE_HPP
#define ADAPTORS_DEFAULT_LOGICAL_FILE_LOGICALFILE_HPP

#include <fstream>
#include <string>

#include <saga/util.hpp>
#include <saga/types.hpp>
#include <saga/adaptors/adaptor.hpp>
#include <saga/adaptors/packages/logical_file_cpi_instance_data.hpp>

#include <impl/engine/proxy.hpp>
#include <impl/packages/logicalfile/logical_file_cpi.hpp>

#include "logicalfile_database.hpp"

///////////////////////////////////////////////////////////////////////////////
namespace logicalfile
{
    ///////////////////////////////////////////////////////////////////////////////
    //  This adaptor implements the functionality of the Saga API "logicalfile".
    //  It defines the functions declared in its base class, logical_file_file_cpi.
    // 
    //  @note This adaptor is written for test purposes only; it is possible that
    //        it does not function exactly like the Saga API dictates.
    //  @see The documentation of the Saga API
    class logical_file_cpi_impl 
        : public saga::adaptors::v1_0::logical_file_cpi<logical_file_cpi_impl>
    {
    private:
        typedef saga::adaptors::v1_0::logical_file_cpi<logical_file_cpi_impl> 
            base_cpi;

        void check_permissions(saga::logical_file::flags flags,
            char const* name, std::string const& lfn);

    public:
        typedef base_cpi::mutex_type mutex_type;
        
        // constructor of the file adaptor 
        logical_file_cpi_impl (proxy* p, cpi_info const& info, 
            saga::ini::ini const& glob_ini, saga::ini::ini const& adap_ini,
            TR1::shared_ptr<saga::adaptor> adaptor);

        // destructor of the file adaptor 
        ~logical_file_cpi_impl  (void);

        // attribute functions
        void sync_attribute_exists(bool&, std::string key);
        void sync_attribute_is_readonly(bool&, std::string key);
        void sync_attribute_is_writable(bool&, std::string key);
        void sync_attribute_is_vector(bool&, std::string key);
        void sync_attribute_is_extended(bool&, std::string key);
        void sync_get_attribute(std::string&, std::string key);
        void sync_set_attribute(saga::void_t&, std::string key, 
            std::string val);
        void sync_get_vector_attribute(std::vector<std::string>&, 
            std::string key);
        void sync_set_vector_attribute(saga::void_t&, std::string, 
            std::vector<std::string>);
        void sync_remove_attribute(saga::void_t&, std::string key);
        void sync_list_attributes(std::vector<std::string>& keys);
        void sync_find_attributes(std::vector<std::string>&, std::string);

        // namespace_entry functions
        void sync_get_url(std::string& url);
        void sync_get_cwd(std::string&);
        void sync_get_name(std::string&);
//         void sync_read_link(std::string&);
        void sync_is_dir(bool&, int);
        void sync_is_entry(bool&, int);
        void sync_is_link(bool&, int);
//         void sync_copy(saga::void_t&, std::string, int);
//         void sync_link(saga::void_t&, std::string, int);
//         void sync_move(saga::void_t&, std::string, int);
        void sync_remove(saga::void_t&, int);
        void sync_close(saga::void_t&, double);

        // logical_file functions
        void sync_list_locations(std::vector<std::string>& locations);
        void sync_add_location(saga::void_t&, std::string location);
        void sync_remove_location(saga::void_t&, std::string location);
        void sync_update_location(saga::void_t&, std::string oldlocation, 
            std::string newlocation);
        void sync_replicate(saga::void_t&, std::string, int);
        
    // This adaptor implements the async functions based on its own synchronous
    // functions for testing purposes only. Since there is no principal need
    // to do so, we allow these to be 'switched off'.
    #if !defined(SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS)
        // attribute functions
        saga::task_type async_attribute_exists(bool&, std::string key);
        saga::task_type async_attribute_is_readonly(bool&, std::string key);
        saga::task_type async_attribute_is_writable(bool&, std::string key);
        saga::task_type async_attribute_is_vector(bool&, std::string key);
        saga::task_type async_attribute_is_extended(bool&, std::string key);
        saga::task_type async_get_attribute(std::string&, std::string key);
        saga::task_type async_set_attribute(saga::void_t&, std::string key, 
            std::string val);
        saga::task_type async_get_vector_attribute(std::vector<std::string>&, 
            std::string key);
        saga::task_type async_set_vector_attribute(saga::void_t&, std::string, 
            std::vector<std::string>);
        saga::task_type async_remove_attribute(saga::void_t&, std::string key);
        saga::task_type async_list_attributes(std::vector<std::string>& keys);
        saga::task_type async_find_attributes(std::vector<std::string>&, 
            std::string);

        // namespace_entry functions
        saga::task_type async_get_url (std::string& url);
        saga::task_type async_get_cwd(std::string&);
        saga::task_type async_get_name(std::string&);
//         saga::task_type async_read_link(std::string&);
        saga::task_type async_is_dir(bool&, int);
        saga::task_type async_is_entry(bool&, int);
        saga::task_type async_is_link(bool&, int);
//         saga::task_type async_copy(saga::void_t&, std::string, int);
//         saga::task_type async_link(saga::void_t&, std::string, int);
//         saga::task_type async_move(saga::void_t&, std::string, int);
        saga::task_type async_remove(saga::void_t&, int);
        saga::task_type async_close(saga::void_t&, double);

        // logical_file functions
        saga::task_type async_list_locations(std::vector<std::string>& locations);
        saga::task_type async_add_location(saga::void_t&, std::string location);
        saga::task_type async_remove_location(saga::void_t&, std::string location);
        saga::task_type async_update_location(saga::void_t&, std::string, 
            std::string);
        saga::task_type async_replicate(saga::void_t&, std::string, int);
    #endif // !SAGA_DEFAULT_FILE_ADAPTOR_NO_ASYNCS

    private:
        database db_;

    // monitoring thread
        boost::thread *thrd_;
        boost::condition *cond_;
        boost::condition cancel_thrd_;
        bool thread_alive_;
        bool cancel_thread_;

    protected:
        void setup_monitoring_thread();
        void monitor_entry();
    };  // class logical_file_cpi_impl

///////////////////////////////////////////////////////////////////////////////
}   // namespace logical_file

#endif // ADAPTORS_DEFAULT_LOGICAL_FILE_LOGICALFILE_HPP

