//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <iostream>
#include <string>

using namespace std;

#include <impl/config.hpp>
#include <impl/engine/cpi.hpp>
#include <impl/engine/engine.hpp>
#include <impl/engine/adaptor_selector.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace saga { namespace impl {

namespace adaptor_selector_helpers {

  using v1_0::preference_type;
  using v1_0::cpi_info;

  /*
    * match_prefs:
    *
    * Searches through the prefs, to see if for all of the items in the
    * needed_prefs there is a matching item in the supplied_prefs.
    *
    * Returns true if so, false if not.
    */
  bool match_prefs(preference_type const& supplied_prefs, 
    preference_type const& needed_prefs) 
  {
    /* create iterators needed for search */
    pair <preference_type::const_iterator, 
          preference_type::const_iterator> found_items;
  
    /* go through all items in needed prefs, and see it there are matching 
      * items in supplied prefs */
    preference_type::const_iterator end = needed_prefs.end ();
    preference_type::const_iterator found;
    preference_type::const_iterator needed;
    
    for ( needed  = needed_prefs.begin ();
          needed != end; 
        ++needed )
    {
      /* get all items with matching keys */
      found_items = supplied_prefs.equal_range (needed->first);
  
      /* go through all found items to see if the value type matches the 
        * needed value type */
      bool match = false;
      
      found = found_items.first;
  
      while ( ! match && found != found_items.second )
      {
        if ( found->second == needed->second )
          match = true;
        
        ++found;
      }
  
      if ( ! match )
      {
        /* no item matched the needed key, value pair */
        return false;
      }
    }
  
    /* for all <key, value> pairs in the needed prefs, a matching
        <key, value> pair in the supplied prefs is found */
    
    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  inline bool 
  match_adaptor_infos(cpi_info const& info, std::string const& op_name, 
    preference_type const& prefs)
  {
  // if the adaptor prefs doesn't match, we don't need to check for the ops
    if ( !match_prefs (prefs, info.get_prefs()) )  
        return false;

  // initialization is always implemented (it's the constructor)
    if (op_name == "__init__")
        return true;
        
    typedef cpi_info::op_list_type op_list_type;
    typedef cpi_info::op_list_type::const_iterator op_list_iterator;

    op_list_type const& ops = info.get_ops();              
    op_list_iterator op_it = ops.find(v1_0::op_info(op_name));
    if (op_it != ops.end()) 
      return match_prefs (prefs, (*op_it).get_prefs());

//     // search for the operation
//     op_list_iterator op_end = ops.end ();
//     for (op_list_iterator op_it = ops.begin (); op_it != op_end; ++op_it)
//     {
//       if ( (op_name == "__init__" || op_name == (*op_it).get_name()) &&  // if opname matches
//            match_prefs (prefs, (*op_it).get_prefs()) )  // if prefs match ops preferences
//       {
//         return true;
//       } // oprefs match
//     } // for ops

    return false;
  } 

} // namespace adaptor_selector_helpers

  /////////////////////////////////////////////////////////////////////////////  
  bool 
  adaptor_selector::match_adaptor(cpi_info const& info, 
    std::string const& ops_cpi_name, std::string const& op_name, 
    preference_type const& prefs)
  {          
    // search adaptors for wanted cpi
    if (info.get_ops_cpi_name() == ops_cpi_name && 
        adaptor_selector_helpers::match_adaptor_infos(info, op_name, prefs))
    {
      return true;
    }
    return false;
  }

  /*
    * select_adaptor:
    *
    * selects a suitable adaptor from the adaptors based on the values defined
    * in the preferences. It returns the maker function of a suitable adaptor.
    *
    * When no suitable adaptor is found, NULL is returned
    */
  adaptor_selector::cpi_info const&
  adaptor_selector::select_adaptor (adaptor_info_list_type const & cpis, 
                                    adaptor_info_list_type const & no_adaptors, 
                                    std::string            const & cpi_name, 
                                    std::string            const & ops_cpi_name, 
                                    std::string            const & op_name,
                                    preference_type        const & prefs) 
  {
    /* find adaptors implementing the supplied interface */
    typedef adaptor_info_list_type::const_iterator adaptor_info_list_iterator;
    
    adaptor_info_list_iterator begin = no_adaptors.begin();
    adaptor_info_list_iterator end = no_adaptors.end();
    
    adaptor_info_list_iterator cpi_end = cpis.end();
    for (adaptor_info_list_iterator cpi_it = cpis.begin(); 
         cpi_it != cpi_end; ++cpi_it)
    {
      // make sure this cpi info belongs to the cpi instance in question
      // if we have to load the adaptor for the constructor (__init__), then we
      // might want to look for the ops_cpi_name as well
      if (cpi_name != (*cpi_it).get_cpi_name() && 
          (op_name != "__init__" || ops_cpi_name != (*cpi_it).get_ops_cpi_name())
         )
      {
          continue;
      }
                    
      // make sure this adaptor is not on the no-no list...
      if ((std::find_if (begin, end, v1_0::same_cpi(*cpi_it)) == end) &&
          match_adaptor(*cpi_it, ops_cpi_name, op_name, prefs))
      {
        return *cpi_it;
      }
    }
    
    SAGA_THROW_VERBATIM(saga::object(), 
        "Could not select any matching adaptor for: " + cpi_name + "::" + 
        op_name, saga::NoAdaptor);
               
    static adaptor_selector::cpi_info dummy;
    return dummy;
  }

  // function to test, whether a given cpi matches the expectations
  bool adaptor_selector::test_adaptor(adaptor_info_list_type const& cpis, 
      v1_0::cpi const* cpi_instance,  
      std::string const& ops_cpi_name, std::string const& op_name,  
      preference_type const& prefs)
  {
      // find adaptors implementing the supplied interface 
      typedef adaptor_info_list_type::const_iterator adaptor_info_list_iterator;
      
      // loop through the list of registered functions for this cpi instance 
      // and check, whether it implements the required function
      cpi_info const& info (cpi_instance->get_adaptor_info());
      saga::uuid cpi_uuid (info.get_cpi_id());
      
      adaptor_info_list_iterator cpi_end = cpis.end();
      for (adaptor_info_list_iterator cpi_it = cpis.begin(); 
           cpi_it != cpi_end; ++cpi_it)
      {
          // make sure this cpi info belongs to the cpi instance in question
          if (cpi_uuid != (*cpi_it).get_cpi_id())
          {
              continue;
          }
          
          // now test, whether preferences match and function is implemented
          if (match_adaptor(*cpi_it, ops_cpi_name, op_name, prefs))
              return true;
      }
      return false;
  }
    
///////////////////////////////////////////////////////////////////////////////
}} // namespace saga::impl


