/*
**  OSSP uuid - Universally Unique Identifier
**  Copyright (c) 2004 Ralf S. Engelschall <rse@engelschall.com>
**  Copyright (c) 2004 The OSSP Project <http://www.ossp.org/>
**
**  This file is part of OSSP uuid, a library for the generation
**  of UUIDs which can found at http://www.ossp.org/pkg/lib/uuid/
**
**  Permission to use, copy, modify, and distribute this software for
**  any purpose with or without fee is hereby granted, provided that
**  the above copyright notice and this permission notice appear in all
**  copies.
**
**  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
**  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
**  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
**  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
**  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
**  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
**  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
**  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
**  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
**  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
**  SUCH DAMAGE.
**
**  saga_uuid.c: library API implementation
*/

/* system headers */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>

#if defined(WIN32) || defined(WIN64)
#include <winsock2.h>
#endif 

/* own headers */
#include "saga_uuid.h"
#include "saga_uuid_vers.h"
#include "saga_uuid_md5.h"
#include "saga_uuid_prng.h"
#include "saga_uuid_mac.h"
#include "saga_uuid_ui64.h"
#include "saga_uuid_str.h"
#include "saga_uuid_bm.h"
#include "saga_uuid_ac.h"

/* maximum number of 100ns ticks of the actual resolution of system clock
   (which in our case is 1us (= 1000ns) because we use gettimeofday(2) */
#define SAGA_UUIDS_PER_TICK 10

/* time offset between SAGA_UUID and Unix Epoch time according to standards.
   (SAGA_UUID UTC base time is October 15, 1582
    Unix UTC base time is January  1, 1970) */
#define SAGA_UUID_TIMEOFFSET "01B21DD213814000"

/* IEEE 802 MAC address encoding/decoding bit fields */
#define IEEE_MAC_MCBIT BM_OCTET(0,0,0,0,0,0,0,1)
#define IEEE_MAC_LOBIT BM_OCTET(0,0,0,0,0,0,1,0)

/* IEEE 802 MAC address octet length */
#define IEEE_MAC_OCTETS 6

/* SAGA_UUID binary representation according to SAGA_UUID standards */
typedef struct {
    saga_uuid_uint32_t  time_low;                  /* bits  0-31 of time field */
    saga_uuid_uint16_t  time_mid;                  /* bits 32-47 of time field */
    saga_uuid_uint16_t  time_hi_and_version;       /* bits 48-59 of time field plus 4 bit version */
    saga_uuid_uint8_t   clock_seq_hi_and_reserved; /* bits  8-13 of clock sequence field plus 2 bit variant */
    saga_uuid_uint8_t   clock_seq_low;             /* bits  0-7  of clock sequence field */
    saga_uuid_uint8_t   node[IEEE_MAC_OCTETS];     /* bits  0-47 of node MAC address */
} saga_uuid_obj_t;

/* abstract data type (ADT) of API */
struct saga_uuid_st {
    saga_uuid_obj_t     obj;                       /* inlined SAGA_UUID object */
    prng_t        *prng;                      /* RPNG sub-object */
    md5_t         *md5;                       /* MD5 sub-object */
    saga_uuid_uint8_t   mac[IEEE_MAC_OCTETS];      /* pre-determined MAC address */
    struct timeval time_last;                 /* last retrieved timestamp */
    unsigned long  time_seq;                  /* last timestamp sequence counter */
};

/* create SAGA_UUID object */
saga_uuid_rc_t saga_uuid_create(saga_uuid_t **saga_uuid)
{
    /* argument sanity check */
    if (saga_uuid == NULL)
        return SAGA_UUID_RC_ARG;

    /* allocate SAGA_UUID object */
    if ((*saga_uuid = (saga_uuid_t *)malloc(sizeof(saga_uuid_t))) == NULL)
        return SAGA_UUID_RC_MEM;

    /* set SAGA_UUID object initially to "Nil SAGA_UUID" */
    saga_uuid_load(*saga_uuid, "nil");

    /* create PRNG and MD5 sub-objects */
    if (prng_create(&(*saga_uuid)->prng) != PRNG_RC_OK)
        return SAGA_UUID_RC_INT;
    if (md5_create(&(*saga_uuid)->md5) != MD5_RC_OK)
        return SAGA_UUID_RC_INT;

    /* resolve MAC address for insertion into node field of SAGA_UUIDs */
    if (!mac_address((unsigned char *)((*saga_uuid)->mac), sizeof((*saga_uuid)->mac))) {
        memset((*saga_uuid)->mac, '\0', sizeof((*saga_uuid)->mac));
        (*saga_uuid)->mac[0] = BM_OCTET(1,0,0,0,0,0,0,0);
    }

    /* initialize time attributes */
    (*saga_uuid)->time_last.tv_sec  = 0;
    (*saga_uuid)->time_last.tv_usec = 0;
    (*saga_uuid)->time_seq = 0;

    return SAGA_UUID_RC_OK;
}

/* destroy SAGA_UUID object */
saga_uuid_rc_t saga_uuid_destroy(saga_uuid_t *saga_uuid)
{
    /* argument sanity check */
    if (saga_uuid == NULL)
        return SAGA_UUID_RC_ARG;

    /* destroy PRNG and MD5 sub-objects */
    prng_destroy(saga_uuid->prng);
    md5_destroy(saga_uuid->md5);

    /* free SAGA_UUID object */
    free(saga_uuid);

    return SAGA_UUID_RC_OK;
}

/* check whether SAGA_UUID object represents "Nil SAGA_UUID" */
saga_uuid_rc_t saga_uuid_isnil(saga_uuid_t *saga_uuid, int *result)
{
    const unsigned char *ucp;
    int i;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || result == NULL)
        return SAGA_UUID_RC_ARG;

    /* a "Nil SAGA_UUID" is defined as all octets zero, so check for this case */
    *result = SAGA_UUID_TRUE;
    for (i = 0, ucp = (unsigned char *)&(saga_uuid->obj); i < SAGA_UUID_LEN_BIN; i++) {
        if (*ucp++ != '\0') {
            *result = SAGA_UUID_FALSE;
            break;
        }
    }

    return SAGA_UUID_RC_OK;
}

/* compare SAGA_UUID objects */
saga_uuid_rc_t saga_uuid_compare(saga_uuid_t *saga_uuid1, saga_uuid_t *saga_uuid2, int *result)
{
    int r;

    /* argument sanity check */
    if (result == NULL)
        return SAGA_UUID_RC_ARG;

    /* convenience macro for setting result */
#   define RESULT(r) \
    do { \
        *result = (r); \
        goto result_exit; \
    } while (0)

    /* special cases: NULL or equal SAGA_UUIDs */
    if (saga_uuid1 == saga_uuid2)
        RESULT(0);
    if (saga_uuid1 == NULL && saga_uuid2 == NULL)
        RESULT(0);
    if (saga_uuid1 == NULL)
        RESULT((saga_uuid_isnil(saga_uuid2, &r), r) ? 0 : -1);
    if (saga_uuid2 == NULL)
        RESULT((saga_uuid_isnil(saga_uuid1, &r), r) ? 0 : 1);

    /* standard cases: regular different SAGA_UUIDs */
    if (saga_uuid1->obj.time_low != saga_uuid2->obj.time_low)
        RESULT((saga_uuid1->obj.time_low < saga_uuid2->obj.time_low) ? -1 : 1);
    if ((r = (int)saga_uuid1->obj.time_mid
           - (int)saga_uuid2->obj.time_mid) != 0)
        RESULT((r < 0) ? -1 : 1);
    if ((r = (int)saga_uuid1->obj.time_hi_and_version
           - (int)saga_uuid2->obj.time_hi_and_version) != 0)
        RESULT((r < 0) ? -1 : 1);
    if ((r = (int)saga_uuid1->obj.clock_seq_hi_and_reserved
           - (int)saga_uuid2->obj.clock_seq_hi_and_reserved) != 0)
        RESULT((r < 0) ? -1 : 1);
    if ((r = (int)saga_uuid1->obj.clock_seq_low
           - (int)saga_uuid2->obj.clock_seq_low) != 0)
        RESULT((r < 0) ? -1 : 1);
    if ((r = memcmp(saga_uuid1->obj.node, saga_uuid2->obj.node, sizeof(saga_uuid1->obj.node))) != 0)
        RESULT((r < 0) ? -1 : 1);

    /* default case: the keys are equal */
    *result = 0;

    result_exit:
    return SAGA_UUID_RC_OK;
}

/* INTERNAL: unpack SAGA_UUID binary presentation into SAGA_UUID object
   (allows in-place operation for internal efficiency!) */
static saga_uuid_rc_t saga_uuid_import_bin(saga_uuid_t *saga_uuid, const void *data_ptr, size_t data_len)
{
    const saga_uuid_uint8_t *in;
    saga_uuid_uint32_t tmp32;
    saga_uuid_uint16_t tmp16;
    unsigned int i;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || data_ptr == NULL || data_len < SAGA_UUID_LEN_BIN)
        return SAGA_UUID_RC_ARG;

    /* treat input data buffer as octet stream */
    in = (const saga_uuid_uint8_t *)data_ptr;

    /* unpack "time_low" field */
    tmp32 = *in++;
    tmp32 = (tmp32 << 8) | *in++;
    tmp32 = (tmp32 << 8) | *in++;
    tmp32 = (tmp32 << 8) | *in++;
    saga_uuid->obj.time_low = tmp32;

    /* unpack "time_mid" field */
    tmp16 = *in++;
    tmp16 = (tmp16 << 8) | *in++;
    saga_uuid->obj.time_mid = tmp16;

    /* unpack "time_hi_and_version" field */
    tmp16 = *in++;
    tmp16 = (tmp16 << 8) | *in++;
    saga_uuid->obj.time_hi_and_version = tmp16;

    /* unpack "clock_seq_hi_and_reserved" field */
    saga_uuid->obj.clock_seq_hi_and_reserved = *in++;

    /* unpack "clock_seq_low" field */
    saga_uuid->obj.clock_seq_low = *in++;

    /* unpack "node" field */
    for (i = 0; i < sizeof(saga_uuid->obj.node); i++)
        saga_uuid->obj.node[i] = *in++;

    return SAGA_UUID_RC_OK;
}

/* INTERNAL: pack SAGA_UUID object into binary representation
   (allows in-place operation for internal efficiency!) */
static saga_uuid_rc_t saga_uuid_export_bin(saga_uuid_t *saga_uuid, void **data_ptr, size_t *data_len)
{
    saga_uuid_uint8_t *out;
    saga_uuid_uint32_t tmp32;
    saga_uuid_uint16_t tmp16;
    unsigned int i;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || data_ptr == NULL)
        return SAGA_UUID_RC_ARG;

    /* optionally allocate octet data buffer */
    if (*data_ptr == NULL) {
        if ((*data_ptr = malloc(sizeof(saga_uuid_t))) == NULL)
            return SAGA_UUID_RC_MEM;
        if (data_len != NULL)
            *data_len = SAGA_UUID_LEN_BIN;
    }
    else {
        if (data_len == NULL)
            return SAGA_UUID_RC_ARG;
        if (*data_len < SAGA_UUID_LEN_BIN)
            return SAGA_UUID_RC_MEM;
        *data_len = SAGA_UUID_LEN_BIN;
    }

    /* treat output data buffer as octet stream */
    out = (saga_uuid_uint8_t *)(*data_ptr);

    /* pack "time_low" field */
    tmp32 = saga_uuid->obj.time_low;
    out[3] = (saga_uuid_uint8_t)(tmp32 & 0xff); tmp32 >>= 8;
    out[2] = (saga_uuid_uint8_t)(tmp32 & 0xff); tmp32 >>= 8;
    out[1] = (saga_uuid_uint8_t)(tmp32 & 0xff); tmp32 >>= 8;
    out[0] = (saga_uuid_uint8_t)(tmp32 & 0xff);

    /* pack "time_mid" field */
    tmp16 = saga_uuid->obj.time_mid;
    out[5] = (saga_uuid_uint8_t)(tmp16 & 0xff); tmp16 >>= 8;
    out[4] = (saga_uuid_uint8_t)(tmp16 & 0xff);

    /* pack "time_hi_and_version" field */
    tmp16 = saga_uuid->obj.time_hi_and_version;
    out[7] = (saga_uuid_uint8_t)(tmp16 & 0xff); tmp16 >>= 8;
    out[6] = (saga_uuid_uint8_t)(tmp16 & 0xff);

    /* pack "clock_seq_hi_and_reserved" field */
    out[8] = saga_uuid->obj.clock_seq_hi_and_reserved;

    /* pack "clock_seq_low" field */
    out[9] = saga_uuid->obj.clock_seq_low;

    /* pack "node" field */
    for (i = 0; i < sizeof(saga_uuid->obj.node); i++)
        out[10+i] = saga_uuid->obj.node[i];

    return SAGA_UUID_RC_OK;
}

/* INTERNAL: check for valid SAGA_UUID string representation syntax */
static int saga_uuid_isstr(const char *str, size_t str_len)
{
    int i;
    const char *cp;

    /* example reference:
       f81d4fae-7dec-11d0-a765-00a0c91e6bf6
       012345678901234567890123456789012345
       0         1         2         3       */
    if (str == NULL)
        return SAGA_UUID_FALSE;
    if (str_len == 0)
        str_len = strlen(str);
    if (str_len < SAGA_UUID_LEN_STR)
        return SAGA_UUID_FALSE;
    for (i = 0, cp = str; i < SAGA_UUID_LEN_STR; i++, cp++) {
        if ((i == 8) || (i == 13) || (i == 18) || (i == 23)) {
            if (*cp == '-')
                continue;
            else
                return SAGA_UUID_FALSE;
        }
        if (!isxdigit((int)(*cp)))
            return SAGA_UUID_FALSE;
    }
    return SAGA_UUID_TRUE;
}

/* INTERNAL: import SAGA_UUID object from string representation */
static saga_uuid_rc_t saga_uuid_import_str(saga_uuid_t *saga_uuid, const void *data_ptr, size_t data_len)
{
    saga_uuid_uint16_t tmp16;
    const char *cp;
    char hexbuf[3];
    const char *str;
    unsigned int i;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || data_ptr == NULL || data_len < SAGA_UUID_LEN_STR)
        return SAGA_UUID_RC_ARG;

    /* check for correct SAGA_UUID string representation syntax */
    str = (const char *)data_ptr;
    if (!saga_uuid_isstr(str, 0))
        return SAGA_UUID_RC_ARG;

    /* parse hex values of "time" parts */
    saga_uuid->obj.time_low            = (saga_uuid_uint32_t)strtoul(str,    NULL, 16);
    saga_uuid->obj.time_mid            = (saga_uuid_uint16_t)strtoul(str+9,  NULL, 16);
    saga_uuid->obj.time_hi_and_version = (saga_uuid_uint16_t)strtoul(str+14, NULL, 16);

    /* parse hex values of "clock" parts */
    tmp16 = (saga_uuid_uint16_t)strtoul(str+19, NULL, 16);
    saga_uuid->obj.clock_seq_low             = (saga_uuid_uint8_t)(tmp16 & 0xff); tmp16 >>= 8;
    saga_uuid->obj.clock_seq_hi_and_reserved = (saga_uuid_uint8_t)(tmp16 & 0xff);

    /* parse hex values of "node" part */
    cp = str+24;
    hexbuf[2] = '\0';
    for (i = 0; i < sizeof(saga_uuid->obj.node); i++) {
        hexbuf[0] = *cp++;
        hexbuf[1] = *cp++;
        saga_uuid->obj.node[i] = (saga_uuid_uint8_t)strtoul(hexbuf, NULL, 16);
    }

    return SAGA_UUID_RC_OK;
}

/* INTERNAL: export SAGA_UUID object to string representation */
static saga_uuid_rc_t saga_uuid_export_str(saga_uuid_t *saga_uuid, void **data_ptr, size_t *data_len)
{
    /* sanity check argument(s) */
    if (saga_uuid == NULL || data_ptr == NULL)
        return SAGA_UUID_RC_ARG;

    /* allocate output buffer */
    if (*data_ptr == NULL) {
        if ((*data_ptr = (void *)malloc(SAGA_UUID_LEN_STR+1)) == NULL)
            return SAGA_UUID_RC_MEM;
        if (data_len != NULL)
            *data_len = SAGA_UUID_LEN_STR+1;
    }
    else {
        if (data_len == NULL)
            return SAGA_UUID_RC_ARG;
        if (*data_len < SAGA_UUID_LEN_STR+1)
            return SAGA_UUID_RC_MEM;
        *data_len = SAGA_UUID_LEN_STR+1;
    }

    /* format SAGA_UUID into string representation */
    str_snprintf((char *)(*data_ptr), SAGA_UUID_LEN_STR+1,
        "%08lx-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
        (unsigned long)saga_uuid->obj.time_low,
        (unsigned int)saga_uuid->obj.time_mid,
        (unsigned int)saga_uuid->obj.time_hi_and_version,
        (unsigned int)saga_uuid->obj.clock_seq_hi_and_reserved,
        (unsigned int)saga_uuid->obj.clock_seq_low,
        (unsigned int)saga_uuid->obj.node[0],
        (unsigned int)saga_uuid->obj.node[1],
        (unsigned int)saga_uuid->obj.node[2],
        (unsigned int)saga_uuid->obj.node[3],
        (unsigned int)saga_uuid->obj.node[4],
        (unsigned int)saga_uuid->obj.node[5]);

    return SAGA_UUID_RC_OK;
}

/* decoding tables */
static struct {
    saga_uuid_uint8_t num;
    const char *desc;
} saga_uuid_dectab_variant[] = {
    { BM_OCTET(0,0,0,0,0,0,0,0), "reserved (NCS backward compatible)" },
    { BM_OCTET(1,0,0,0,0,0,0,0), "DCE 1.1, ISO/IEC 11578:1996" },
    { BM_OCTET(1,1,0,0,0,0,0,0), "reserved (Microsoft GUID)" },
    { BM_OCTET(1,1,1,0,0,0,0,0), "reserved (future use)" }
};
static struct {
    int num;
    const char *desc;
} saga_uuid_dectab_version[] = {
    { 1, "time and node based" },
    { 3, "name based" },
    { 4, "random data based" }
};

/* INTERNAL: dump SAGA_UUID object as descriptive text */
static saga_uuid_rc_t saga_uuid_export_txt(saga_uuid_t *saga_uuid, void **data_ptr, size_t *data_len)
{
    saga_uuid_rc_t      rc;
    char              **out;
    char               *out_ptr;
    size_t              out_len;
    const char         *version;
    const char         *variant;
    char               *content;
    int                isnil;
    saga_uuid_uint8_t  tmp8;
    saga_uuid_uint16_t tmp16;
    saga_uuid_uint32_t tmp32;
    saga_uuid_uint8_t  tmp_bin[SAGA_UUID_LEN_BIN];
    char               tmp_str[SAGA_UUID_LEN_STR+1];
    void              *tmp_ptr;
    size_t             tmp_len;
    ui64_t             t;
    ui64_t             t_offset;
    int                t_nsec;
    int                t_usec;
    time_t             t_sec;
    char               t_buf[19+1]; /* YYYY-MM-DD HH:MM:SS */
    struct tm         *tm;
    unsigned int       i;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || data_ptr == NULL)
        return SAGA_UUID_RC_ARG;

    /* initialize output buffer */
    out_ptr = NULL;
    out = &out_ptr;

    /* check for special case of "Nil SAGA_UUID" */
    if ((rc = saga_uuid_isnil(saga_uuid, &isnil)) != SAGA_UUID_RC_OK)
        return rc;

    /* decode into string representation */
    tmp_ptr = (void *)&tmp_str;
    tmp_len = sizeof(tmp_str);
    if ((rc = saga_uuid_export(saga_uuid, SAGA_UUID_FMT_STR, &tmp_ptr, &tmp_len)) != SAGA_UUID_RC_OK)
        return rc;
    str_rsprintf(out, "SAGA_UUID:    %s\n", tmp_str);

    /* decode SAGA_UUID variant */
    tmp8 = saga_uuid->obj.clock_seq_hi_and_reserved;
    if (isnil)
        variant = "n.a.";
    else {
        variant = "unknown";
        for (i = 7; i >= 0; i--) {
            if ((tmp8 & BM_BIT(i,1)) == 0) {
                tmp8 &= ~BM_MASK(i,0);
                break;
            }
        }
        
        for ( i = 0; i < sizeof(saga_uuid_dectab_variant)/sizeof(saga_uuid_dectab_variant[0]); i++) 
        {
            if (saga_uuid_dectab_variant[i].num == tmp8) {
                variant = saga_uuid_dectab_variant[i].desc;
                break;
            }
        }
    }
    str_rsprintf(out, "variant: %s\n", variant);

    /* decode SAGA_UUID version */
    tmp16 = (BM_SHR(saga_uuid->obj.time_hi_and_version, 12) & BM_MASK(3,0));
    if (isnil)
        version = "n.a.";
    else {
        version = "unknown";
        for (i = 0; i < sizeof(saga_uuid_dectab_version)/sizeof(saga_uuid_dectab_version[0]); i++) {
            if (saga_uuid_dectab_version[i].num == (int)tmp16) {
                version = saga_uuid_dectab_version[i].desc;
                break;
            }
        }
    }
    str_rsprintf(out, "version: %d (%s)\n", (int)tmp16, version);

    /*
     * decode SAGA_UUID content
     */

    if (tmp8 == BM_OCTET(1,0,0,0,0,0,0,0) && tmp16 == 1) {
        /* decode DCE 1.1 version 1 SAGA_UUID */

        /* decode system time */
        t = ui64_rol(ui64_n2i((unsigned long)(saga_uuid->obj.time_hi_and_version & BM_MASK(11,0))), 48, NULL),
        t = ui64_or(t, ui64_rol(ui64_n2i((unsigned long)(saga_uuid->obj.time_mid)), 32, NULL));
        t = ui64_or(t, ui64_n2i((unsigned long)(saga_uuid->obj.time_low)));
        t_offset = ui64_s2i(SAGA_UUID_TIMEOFFSET, NULL, 16);
        t = ui64_sub(t, t_offset, NULL);
        t = ui64_divn(t, 10, &t_nsec);
        t = ui64_divn(t, 1000000, &t_usec);
        t_sec = (time_t)ui64_i2n(t);
        tm = gmtime(&t_sec);
        strftime(t_buf, sizeof(t_buf), "%Y-%m-%d %H:%M:%S", tm);
        str_rsprintf(out, "content: time:  %s.%06d.%d UTC\n", t_buf, t_usec, t_nsec);

        /* decode clock sequence */
        tmp32 = ((saga_uuid->obj.clock_seq_hi_and_reserved & BM_MASK(5,0)) << 8)
                + saga_uuid->obj.clock_seq_low;
        str_rsprintf(out, "         clock: %ld (usually random)\n", (long)tmp32);

        /* decode node MAC address */
        str_rsprintf(out, "         node:  %02x:%02x:%02x:%02x:%02x:%02x (%s %s)\n",
            (unsigned int)saga_uuid->obj.node[0],
            (unsigned int)saga_uuid->obj.node[1],
            (unsigned int)saga_uuid->obj.node[2],
            (unsigned int)saga_uuid->obj.node[3],
            (unsigned int)saga_uuid->obj.node[4],
            (unsigned int)saga_uuid->obj.node[5],
            (saga_uuid->obj.node[0] & IEEE_MAC_LOBIT ? "local" : "global"),
            (saga_uuid->obj.node[0] & IEEE_MAC_MCBIT ? "multicast" : "unicast"));
    }
    else {
        /* decode anything else as hexadecimal byte-string only */

        /* determine annotational hint */
        content = "not decipherable, because unknown SAGA_UUID version";
        if (isnil)
            content = "special case of DCE 1.1 Nil SAGA_UUID";
        else if (tmp16 == 3)
            content = "not decipherable, because message digest only";
        else if (tmp16 == 4)
            content = "no semantics, because random data only";

        /* pack SAGA_UUID into binary representation */
        tmp_ptr = (void *)&tmp_bin;
        tmp_len = sizeof(tmp_bin);
        if ((rc = saga_uuid_export(saga_uuid, SAGA_UUID_FMT_BIN, &tmp_ptr, &tmp_len)) !=
            SAGA_UUID_RC_OK)
            return rc;

        /* mask out version and variant parts */
        tmp_bin[6] &= BM_MASK(3,0);
        tmp_bin[8] &= BM_MASK(5,0);

        /* dump as colon-seperated hexadecimal byte-string */
        str_rsprintf(out,
            "content: %02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X\n"
            "         (%s)\n",
            (unsigned int)tmp_bin[0],  (unsigned int)tmp_bin[1],  (unsigned int)tmp_bin[2],
            (unsigned int)tmp_bin[3],  (unsigned int)tmp_bin[4],  (unsigned int)tmp_bin[5],
            (unsigned int)tmp_bin[6],  (unsigned int)tmp_bin[7],  (unsigned int)tmp_bin[8],
            (unsigned int)tmp_bin[9],  (unsigned int)tmp_bin[10], (unsigned int)tmp_bin[11],
            (unsigned int)tmp_bin[12], (unsigned int)tmp_bin[13], (unsigned int)tmp_bin[14],
            (unsigned int)tmp_bin[15], content);
    }

    /* provide result */
    out_len = strlen(out_ptr)+1;
    if (*data_ptr == NULL) {
        *data_ptr = (void *)out_ptr;
        if (data_len != NULL)
            *data_len = out_len;
    }
    else {
        if (data_len == NULL)
            return SAGA_UUID_RC_ARG;
        if (*data_len < out_len)
            return SAGA_UUID_RC_MEM;
        memcpy(*data_ptr, out_ptr, out_len);
    }

    return SAGA_UUID_RC_OK;
}

/* SAGA_UUID importing */
saga_uuid_rc_t saga_uuid_import(saga_uuid_t *saga_uuid, saga_uuid_fmt_t fmt, const void *data_ptr, size_t data_len)
{
    saga_uuid_rc_t rc;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || data_ptr == NULL)
        return SAGA_UUID_RC_ARG;

    /* dispatch into format-specific functions */
    switch (fmt) {
        case SAGA_UUID_FMT_BIN: rc = saga_uuid_import_bin(saga_uuid, data_ptr, data_len); break;
        case SAGA_UUID_FMT_STR: rc = saga_uuid_import_str(saga_uuid, data_ptr, data_len); break;
        case SAGA_UUID_FMT_TXT: rc = SAGA_UUID_RC_IMP; /* not implemented */ break;
        default:           rc = SAGA_UUID_RC_ARG;
    }

    return rc;
}

/* SAGA_UUID exporting */
saga_uuid_rc_t saga_uuid_export(saga_uuid_t *saga_uuid, saga_uuid_fmt_t fmt, void **data_ptr, size_t *data_len)
{
    saga_uuid_rc_t rc;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || data_ptr == NULL)
        return SAGA_UUID_RC_ARG;

    /* dispatch into format-specific functions */
    switch (fmt) {
        case SAGA_UUID_FMT_BIN: rc = saga_uuid_export_bin(saga_uuid, data_ptr, data_len); break;
        case SAGA_UUID_FMT_STR: rc = saga_uuid_export_str(saga_uuid, data_ptr, data_len); break;
        case SAGA_UUID_FMT_TXT: rc = saga_uuid_export_txt(saga_uuid, data_ptr, data_len); break;
        default:           rc = SAGA_UUID_RC_ARG;
    }

    return rc;
}

/* INTERNAL: brand SAGA_UUID with version and variant */
static void saga_uuid_brand(saga_uuid_t *saga_uuid, int version)
{
    /* set version (as given) */
    saga_uuid->obj.time_hi_and_version &= BM_MASK(11,0);
    saga_uuid->obj.time_hi_and_version |= BM_SHL((saga_uuid_uint16_t)version, 12);

    /* set variant (always DCE 1.1 only) */
    saga_uuid->obj.clock_seq_hi_and_reserved &= BM_MASK(5,0);
    saga_uuid->obj.clock_seq_hi_and_reserved |= BM_SHL(0x02, 6);
    return;
}

/* INTERNAL: generate SAGA_UUID version 1: time, clock and node based */
static saga_uuid_rc_t saga_uuid_make_v1(saga_uuid_t *saga_uuid, unsigned int mode, va_list ap)
{
    struct timeval time_now;
#ifdef HAVE_NANOSLEEP
    struct timespec ts;
#elif !((_WIN32_WINNT >= 0x0400) || (_WIN32_WINDOWS > 0x0400))
    struct timeval tv;
#endif
    ui64_t t;
    ui64_t offset;
    ui64_t ov;
    saga_uuid_uint16_t clck;

    /*
     *  GENERATE TIME
     */

    /* determine current system time and sequence counter */
    while (1) {
        /* determine current system time */
        if (gettimeofday(&time_now, NULL) == -1)
            return SAGA_UUID_RC_SYS;

        /* check whether system time changed since last retrieve */
        if (!(   time_now.tv_sec  == saga_uuid->time_last.tv_sec
              && time_now.tv_usec == saga_uuid->time_last.tv_usec)) {
            /* reset time sequence counter and continue */
            saga_uuid->time_seq = 0;
            break;
        }

        /* until we are out of SAGA_UUIDs per tick, increment
           the time/tick sequence counter and continue */
        if (saga_uuid->time_seq < SAGA_UUIDS_PER_TICK) {
            saga_uuid->time_seq++;
            break;
        }

        /* stall the SAGA_UUID generation until the system clock (which
           has a gettimeofday(2) resolution of 1us) catches up */
#ifdef HAVE_NANOSLEEP
        /* sleep for 500ns (1/2us) */
        ts.tv_sec  = 0;
        ts.tv_nsec = 500;
        nanosleep(&ts, NULL);
#elif (_WIN32_WINNT >= 0x0400) || (_WIN32_WINDOWS > 0x0400)
        usleep(1);
#else
        /* sleep for 1000ns (1us) */
        tv.tv_sec  = 0;
        tv.tv_usec = 1;
        select(0, NULL, NULL, NULL, &tv);
#endif
    }

    /* convert from timeval (sec,usec) to OSSP ui64 (100*nsec) format */
    t = ui64_n2i(time_now.tv_sec);
    t = ui64_muln(t, 1000000, NULL);
    t = ui64_addn(t, time_now.tv_usec, NULL);
    t = ui64_muln(t, 10, NULL);

    /* adjust for offset between SAGA_UUID and Unix Epoch time */
    offset = ui64_s2i(SAGA_UUID_TIMEOFFSET, NULL, 16);
    t = ui64_add(t, offset, NULL);

    /* compensate for low resolution system clock by adding
       the time/tick sequence counter */
    if (saga_uuid->time_seq > 0)
        t = ui64_addn(t, saga_uuid->time_seq, NULL);

    /* store the 60 LSB of the time in the SAGA_UUID */
    t = ui64_rol(t, 16, &ov);
    saga_uuid->obj.time_hi_and_version =
        (saga_uuid_uint16_t)(ui64_i2n(ov) & 0x00000fff); /* 12 of 16 bit only! */
    t = ui64_rol(t, 16, &ov);
    saga_uuid->obj.time_mid =
        (saga_uuid_uint16_t)(ui64_i2n(ov) & 0x0000ffff); /* all 16 bit */
    t = ui64_rol(t, 32, &ov);
    saga_uuid->obj.time_low =
        (saga_uuid_uint32_t)(ui64_i2n(ov) & 0xffffffff); /* all 32 bit */

    /*
     *  GENERATE CLOCK
     */

    /* retrieve current clock sequence */
    clck = ((saga_uuid->obj.clock_seq_hi_and_reserved & BM_MASK(5,0)) << 8)
           + saga_uuid->obj.clock_seq_low;

    /* generate new random clock sequence (initially or if the
       time has stepped backwards) or else just increase it */
    if (   clck == 0
        || (   time_now.tv_sec < saga_uuid->time_last.tv_sec
            || (   time_now.tv_sec == saga_uuid->time_last.tv_sec
                && time_now.tv_usec < saga_uuid->time_last.tv_usec)))
        prng_data(saga_uuid->prng, (void *)&clck, sizeof(clck));
    else
        clck++;
    clck %= BM_POW2(14);

    /* store back new clock sequence */
    saga_uuid->obj.clock_seq_hi_and_reserved =
        (saga_uuid->obj.clock_seq_hi_and_reserved & BM_MASK(7,6))
        | (saga_uuid_uint8_t)((clck >> 8) & 0xff);
    saga_uuid->obj.clock_seq_low =
        (saga_uuid_uint8_t)(clck & 0xff);

    /*
     *  GENERATE NODE
     */

    if ((mode & SAGA_UUID_MAKE_MC) || (saga_uuid->mac[0] & BM_OCTET(1,0,0,0,0,0,0,0))) {
        /* generate random IEEE 802 local multicast MAC address */
        prng_data(saga_uuid->prng, (void *)&(saga_uuid->obj.node), sizeof(saga_uuid->obj.node));
        saga_uuid->obj.node[0] |= IEEE_MAC_MCBIT;
        saga_uuid->obj.node[0] |= IEEE_MAC_LOBIT;
    }
    else {
        /* use real regular MAC address */
        memcpy(saga_uuid->obj.node, saga_uuid->mac, sizeof(saga_uuid->mac));
    }

    /*
     *  FINISH
     */

    /* remember current system time for next iteration */
    saga_uuid->time_last.tv_sec  = time_now.tv_sec;
    saga_uuid->time_last.tv_usec = time_now.tv_usec;

    /* brand with version and variant */
    saga_uuid_brand(saga_uuid, 1);

    return SAGA_UUID_RC_OK;
}

/* INTERNAL: pre-defined SAGA_UUID values.
   (defined as network byte ordered octet stream) */
static struct {
    char *name;
    saga_uuid_uint8_t saga_uuid[SAGA_UUID_LEN_BIN];
} saga_uuid_value_table[] = {
    { "nil",     /* 00000000-0000-0000-0000-000000000000 ("Nil SAGA_UUID") */
      { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 } },
    { "ns:DNS",  /* 6ba7b810-9dad-11d1-80b4-00c04fd430c8 (see draft-leach-saga_uuids-guids-01.txt) */
      { 0x6b,0xa7,0xb8,0x10,0x9d,0xad,0x11,0xd1,0x80,0xb4,0x00,0xc0,0x4f,0xd4,0x30,0xc8 } },
    { "ns:URL",  /* 6ba7b811-9dad-11d1-80b4-00c04fd430c8 (see draft-leach-saga_uuids-guids-01.txt) */
      { 0x6b,0xa7,0xb8,0x11,0x9d,0xad,0x11,0xd1,0x80,0xb4,0x00,0xc0,0x4f,0xd4,0x30,0xc8 } },
    { "ns:OID",  /* 6ba7b812-9dad-11d1-80b4-00c04fd430c8 (see draft-leach-saga_uuids-guids-01.txt) */
      { 0x6b,0xa7,0xb8,0x12,0x9d,0xad,0x11,0xd1,0x80,0xb4,0x00,0xc0,0x4f,0xd4,0x30,0xc8 } },
    { "ns:X500", /* 6ba7b814-9dad-11d1-80b4-00c04fd430c8 (see draft-leach-saga_uuids-guids-01.txt) */
      { 0x6b,0xa7,0xb8,0x14,0x9d,0xad,0x11,0xd1,0x80,0xb4,0x00,0xc0,0x4f,0xd4,0x30,0xc8 } }
};

/* load SAGA_UUID object with pre-defined value */
saga_uuid_rc_t saga_uuid_load(saga_uuid_t *saga_uuid, const char *name)
{
    saga_uuid_uint8_t *saga_uuid_octets;
    saga_uuid_rc_t rc;
    unsigned int i;

    /* sanity check argument(s) */
    if (saga_uuid == NULL || name == NULL)
        return SAGA_UUID_RC_ARG;

    /* search for SAGA_UUID in table */
    saga_uuid_octets = NULL;
    for (i = 0; i < sizeof(saga_uuid_value_table)/sizeof(saga_uuid_value_table[0]); i++) {
         if (strcmp(saga_uuid_value_table[i].name, name) == 0) {
             saga_uuid_octets = saga_uuid_value_table[i].saga_uuid;
             break;
         }
    }
    if (saga_uuid_octets == NULL)
        return SAGA_UUID_RC_ARG;

    /* import value into SAGA_UUID object */
    if ((rc = saga_uuid_import(saga_uuid, SAGA_UUID_FMT_BIN, saga_uuid_octets,
                          SAGA_UUID_LEN_BIN)) != SAGA_UUID_RC_OK)
        return rc;

    return SAGA_UUID_RC_OK;
}

/* INTERNAL: generate SAGA_UUID version 3: name based */
static saga_uuid_rc_t saga_uuid_make_v3(saga_uuid_t *saga_uuid, unsigned int mode, va_list ap)
{
    char *str;
    saga_uuid_t *saga_uuid_ns;
    saga_uuid_uint8_t saga_uuid_buf[SAGA_UUID_LEN_BIN];
    void *saga_uuid_ptr;
    size_t saga_uuid_len;

    /* determine namespace SAGA_UUID and name string arguments */
    if ((saga_uuid_ns = (saga_uuid_t *)va_arg(ap, void *)) == NULL)
        return SAGA_UUID_RC_ARG;
    if ((str = (char *)va_arg(ap, char *)) == NULL)
        return SAGA_UUID_RC_ARG;

    /* initialize MD5 context */
    if (md5_init(saga_uuid->md5) != MD5_RC_OK)
        return SAGA_UUID_RC_MEM;

    /* load the namespace SAGA_UUID into MD5 context */
    saga_uuid_ptr = (void *)&saga_uuid_buf;
    saga_uuid_len = sizeof(saga_uuid_buf);
    saga_uuid_export(saga_uuid_ns, SAGA_UUID_FMT_BIN, &saga_uuid_ptr, &saga_uuid_len);
    md5_update(saga_uuid->md5, saga_uuid_buf, saga_uuid_len);

    /* load the argument name string into MD5 context */
    md5_update(saga_uuid->md5, str, strlen(str));

    /* store MD5 result into SAGA_UUID
       (requires MD5_LEN_BIN space, SAGA_UUID_LEN_BIN space is available,
       and both are equal in size, so we are safe!) */
    saga_uuid_ptr = (void *)&(saga_uuid->obj);
    md5_store(saga_uuid->md5, &saga_uuid_ptr, NULL);

    /* fulfill requirement of standard and convert SAGA_UUID data into
       local/host byte order (this uses fact that saga_uuid_import_bin() is
       able to operate in-place!) */
    saga_uuid_import(saga_uuid, SAGA_UUID_FMT_BIN, (void *)&(saga_uuid->obj), SAGA_UUID_LEN_BIN);

    /* brand SAGA_UUID with version and variant */
    saga_uuid_brand(saga_uuid, 3);

    return SAGA_UUID_RC_OK;
}

/* INTERNAL: generate SAGA_UUID version 4: random number based */
static saga_uuid_rc_t saga_uuid_make_v4(saga_uuid_t *saga_uuid, unsigned int mode, va_list ap)
{
    /* fill SAGA_UUID with random data */
    prng_data(saga_uuid->prng, (void *)&(saga_uuid->obj), sizeof(saga_uuid->obj));

    /* brand SAGA_UUID with version and variant */
    saga_uuid_brand(saga_uuid, 4);

    return SAGA_UUID_RC_OK;
}

#if defined(WIN32) || defined(WIN64)
static saga_uuid_rc_t saga_uuid_make_system(saga_uuid_t *saga_uuid, unsigned int mode, va_list ap)
{
    UUID guid;
    UuidCreate(&guid);
    
    saga_uuid->obj.time_low = guid.Data1;
    saga_uuid->obj.time_mid = guid.Data2;
    saga_uuid->obj.time_hi_and_version = guid.Data3;
    saga_uuid->obj.clock_seq_hi_and_reserved = guid.Data4[0];
    saga_uuid->obj.clock_seq_low = guid.Data4[1];
    memcpy(&saga_uuid->obj.node, &guid.Data4[2], IEEE_MAC_OCTETS);
    
    return SAGA_UUID_RC_OK;
}
#endif

/* generate SAGA_UUID */
saga_uuid_rc_t saga_uuid_make(saga_uuid_t *saga_uuid, unsigned int mode, ...)
{
    va_list ap;
    saga_uuid_rc_t rc;

    /* sanity check argument(s) */
    if (saga_uuid == NULL)
        return SAGA_UUID_RC_ARG;

    /* dispatch into version dependent generation functions */
    va_start(ap, mode);
    if (mode & SAGA_UUID_MAKE_V1)
        rc = saga_uuid_make_v1(saga_uuid, mode, ap);
    else if (mode & SAGA_UUID_MAKE_V3)
        rc = saga_uuid_make_v3(saga_uuid, mode, ap);
    else if (mode & SAGA_UUID_MAKE_V4)
        rc = saga_uuid_make_v4(saga_uuid, mode, ap);
#if defined(WIN32) || defined(WIN64)
    else if (mode & SAGA_UUID_MAKE_SYSTEM)
        rc = saga_uuid_make_system(saga_uuid, mode, ap);
#endif
    else
        rc = SAGA_UUID_RC_ARG;
    va_end(ap);

    return rc;
}

/* translate SAGA_UUID API error code into corresponding error string */
char const *saga_uuid_error(saga_uuid_rc_t rc)
{
    char const *str;
    switch (rc) {
        case SAGA_UUID_RC_OK:  str = "everything ok";    break;
        case SAGA_UUID_RC_ARG: str = "invalid argument"; break;
        case SAGA_UUID_RC_MEM: str = "out of memory";    break;
        case SAGA_UUID_RC_SYS: str = "system error";     break;
        case SAGA_UUID_RC_INT: str = "internal error";   break;
        case SAGA_UUID_RC_IMP: str = "not implemented";  break;
        default:               str = "unknown error";    break;
    }
    return str;
}

/* OSSP saga_uuid version (link-time information) */
unsigned long saga_uuid_version(void)
{
    return (unsigned long)(_SAGA_UUID_VERSION);
}

