/*
**  OSSP uuid - Universally Unique Identifier
**  Copyright (c) 2004 Ralf S. Engelschall <rse@engelschall.com>
**  Copyright (c) 2004 The OSSP Project <http://www.ossp.org/>
**
**  This file is part of OSSP uuid, a library for the generation
**  of UUIDs which can found at http://www.ossp.org/pkg/lib/uuid/
**
**  Permission to use, copy, modify, and distribute this software for
**  any purpose with or without fee is hereby granted, provided that
**  the above copyright notice and this permission notice appear in all
**  copies.
**
**  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
**  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
**  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
**  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
**  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
**  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
**  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
**  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
**  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
**  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
**  SUCH DAMAGE.
**
**  saga_uuid.h: library API definition
*/

#ifndef __SAGA_UUID_H__
#define __SAGA_UUID_H__

/* required system headers */
#include <string.h>
#include <saga/util.hpp>

/* minimum C++ support */
#ifdef __cplusplus
#define DECLARATION_BEGIN extern "C" {
#define DECLARATION_END   }
#else
#define DECLARATION_BEGIN
#define DECLARATION_END
#endif

DECLARATION_BEGIN

/* OSSP saga_uuid version (compile-time information) */
#define SAGA_UUID_VERSION  0x100300

/* encoding octet stream lengths */
#define SAGA_UUID_LEN_BIN  (128 /*bit*/ / 8 /*bytes*/)
#define SAGA_UUID_LEN_STR  (128 /*bit*/ / 4 /*nibbles*/ + 4 /*hyphens*/)

/* API return codes */
typedef enum {
    SAGA_UUID_RC_OK   = 0,        /* everything ok    */
    SAGA_UUID_RC_ARG  = 1,        /* invalid argument */
    SAGA_UUID_RC_MEM  = 2,        /* out of memory    */
    SAGA_UUID_RC_SYS  = 3,        /* system error     */
    SAGA_UUID_RC_INT  = 4,        /* internal error   */
    SAGA_UUID_RC_IMP  = 5         /* not implemented  */
} saga_uuid_rc_t;

/* SAGA_UUID make modes */
enum {
    SAGA_UUID_MAKE_V1 = (1 << 0), /* DCE 1.1 v1 SAGA_UUID */
    SAGA_UUID_MAKE_V3 = (1 << 1), /* DCE 1.1 v3 SAGA_UUID */
    SAGA_UUID_MAKE_V4 = (1 << 2), /* DCE 1.1 v4 SAGA_UUID */
#if defined(WIN32) || defined(WIN64)
    SAGA_UUID_MAKE_SYSTEM = (1 << 3), /* rely on the operating system to generate an uuid */
#else
    SAGA_UUID_MAKE_SYSTEM = SAGA_UUID_MAKE_V1, 
#endif
    SAGA_UUID_MAKE_MC = (1 << 4)  /* enforce multi-cast MAC address */
};

/* SAGA_UUID import/export formats */
typedef enum {
    SAGA_UUID_FMT_BIN = 0,        /* binary representation (import/export) */
    SAGA_UUID_FMT_STR = 1,        /* string representation (import/export) */
    SAGA_UUID_FMT_TXT = 2         /* textual description   (export only)   */
} saga_uuid_fmt_t;

/* SAGA_UUID abstract data type */
struct saga_uuid_st;
typedef struct saga_uuid_st saga_uuid_t;

/* SAGA_UUID object handling */
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_create   (saga_uuid_t **_saga_uuid);
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_destroy  (saga_uuid_t  *_saga_uuid);

/* SAGA_UUID generation */
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_load     (saga_uuid_t  *_saga_uuid, const char *_name);
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_make     (saga_uuid_t  *_saga_uuid, unsigned int _mode, ...);

/* SAGA_UUID comparison */
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_isnil    (saga_uuid_t  *_saga_uuid,                 int *_result);
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_compare  (saga_uuid_t  *_saga_uuid, saga_uuid_t
                                            *_saga_uuid2, int *_result);

/* SAGA_UUID import/export */
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_import   (saga_uuid_t  *_saga_uuid, saga_uuid_fmt_t _fmt, const void  *_data_ptr, size_t  _data_len);
SAGA_EXPORT extern saga_uuid_rc_t     saga_uuid_export   (saga_uuid_t  *_saga_uuid, saga_uuid_fmt_t _fmt,       void **_data_ptr, size_t *_data_len);

/* library utilities */
SAGA_EXPORT extern char const*   saga_uuid_error    (saga_uuid_rc_t _rc);
SAGA_EXPORT extern unsigned long saga_uuid_version  (void);

DECLARATION_END

#endif /* __SAGA_UUID_H__ */

