//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <saga/exception.hpp>
#include <saga/object.hpp>
#include <saga/task.hpp>
#include <saga/attribute.hpp>
#include <saga/context.hpp>
#include <saga/adaptors/attribute.hpp>
#include <saga/adaptors/attribute_cpi_wrapper.hpp> 

#include <impl/exception.hpp>
#include <impl/engine/object.hpp>
#include <impl/engine/attribute.hpp> 
#include <impl/engine/context_base.hpp> 
#include <impl/engine/proxy.hpp> 

#include <saga/detail/attribute_impl.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace saga { namespace adaptors {

    attribute::attribute (saga::object rhs)
    :   impl_(rhs.get_impl().get()), 
        attr_(rhs.get_impl()->get_attributes()),
        must_delete(false)
    {
    }

    attribute::attribute (saga::impl::v1_0::attribute_cpi *cpi_instance)
    :   impl_(cpi_instance->get_proxy()),
        attr_(new attribute_cpi_wrapper(cpi_instance)),
        must_delete(true)
    {
    }

    attribute::~attribute (void)
    {
        if (must_delete)
            delete attr_;
    }

    ///////////////////////////////////////////////////////////////////////////
    struct set_attribute_priv
    {
        template <typename Derived>
        static task_type 
        call (Derived& this_, saga::void_t& ret, 
            std::string const& key, std::string const& val, bool sync)
        {
            return this_.get_attr()->set_attribute (ret, key, val, sync);
        }
    };
      
    template <typename Tag>
    inline task_type
    attribute::set_attributepriv (saga::void_t& ret, 
        std::string const& key, std::string const& val, Tag) 
    {
        return detail::dispatch_priv<Tag>::
            template call<set_attribute_priv>(derived(), ret, key, val);
    }

    template task_type 
    attribute::set_attributepriv<saga::task_base::Sync>(saga::void_t& ret, 
        std::string const& key, std::string const& val, saga::task_base::Sync);
    template task_type 
    attribute::set_attributepriv<saga::task_base::ASync>(saga::void_t& ret, 
        std::string const& key, std::string const& val, saga::task_base::ASync);
    template task_type 
    attribute::set_attributepriv<saga::task_base::Task>(saga::void_t& ret, 
        std::string const& key, std::string const& val, saga::task_base::Task);
    
    ///////////////////////////////////////////////////////////////////////////
    struct set_vector_attribute_priv
    {
        template <typename Derived>
        static task_type
        call (Derived& this_, saga::void_t& ret, std::string const& key, 
            typename attribute::strvec_type const& val, bool sync)
        { 
            return this_.get_attr()->set_vector_attribute (ret, key, val, sync);
        }
     };
     
    template <typename Tag>
    inline task_type
    attribute::set_vector_attributepriv (saga::void_t& ret, 
        std::string const& key, strvec_type const& val, Tag) 
    {
        return detail::dispatch_priv<Tag>::
            template call<set_vector_attribute_priv>(derived(), ret, key, val);
    }

    template task_type 
    attribute::set_vector_attributepriv<saga::task_base::Sync>(saga::void_t& ret, 
        std::string const& key, strvec_type const& val, saga::task_base::Sync);
    template task_type 
    attribute::set_vector_attributepriv<saga::task_base::ASync>(saga::void_t& ret, 
        std::string const& key, strvec_type const& val, saga::task_base::ASync);
    template task_type 
    attribute::set_vector_attributepriv<saga::task_base::Task>(saga::void_t& ret, 
        std::string const& key, strvec_type const& val, saga::task_base::Task);

    ///////////////////////////////////////////////////////////////////////////
    struct remove_attribute_priv
    {
        template <typename Derived>
        static task_type
        call (Derived& this_, saga::void_t& ret, std::string const& key, bool sync)
        {
            if (!this_.attribute_exists(key))
            {
                // DoesNotExist
                SAGA_THROW_VERBATIM(this_.get_target_object(), 
                    "attribute '" + key + "' does not exist", 
                    saga::DoesNotExist);
            }
            return this_.get_attr()->remove_attribute(ret, key, sync);
        }
    };

    template <typename Tag>
    inline task_type
    attribute::remove_attributepriv (saga::void_t& ret, 
        std::string const& key, Tag) 
    {
        return detail::dispatch_priv<Tag>::
            template call<remove_attribute_priv>(derived(), ret, key);
    }

    template task_type 
    attribute::remove_attributepriv<saga::task_base::Sync>(saga::void_t& ret, 
        std::string const& key, saga::task_base::Sync);
    template task_type 
    attribute::remove_attributepriv<saga::task_base::ASync>(saga::void_t& ret, 
        std::string const& key, saga::task_base::ASync);
    template task_type 
    attribute::remove_attributepriv<saga::task_base::Task>(saga::void_t& ret, 
        std::string const& key, saga::task_base::Task);

}   // namespace adaptors

///////////////////////////////////////////////////////////////////////////
// implement the attribute functions (we need to explicitly specialize 
// the template because the functions are not implemented inline)
template struct detail::attribute <adaptors::attribute>;

template struct detail::attribute_priv<adaptors::attribute, saga::task_base::Sync>;
template struct detail::attribute_priv<adaptors::attribute, saga::task_base::Async>;
template struct detail::attribute_priv<adaptors::attribute, saga::task_base::Task>;
  
///////////////////////////////////////////////////////////////////////////////
}  // namespace saga

