//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
//  Copyright (c) 2007 Ole Weidner (oweidner@cct.lsu.edu)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef SAGA_ADAPTORS_ATTRIBUTE_HPP
#define SAGA_ADAPTORS_ATTRIBUTE_HPP

// include stl
#include <vector>
#include <string>

// include dependent spec sections
#include <saga/util.hpp>
#include <saga/base.hpp>
#include <saga/attribute.hpp>
#include <saga/adaptors/attribute_cpi.hpp>

// suppress warnings about dependent classes not being exported from the dll
#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable : 4251 4231 4660)
#endif

///////////////////////////////////////////////////////////////////////////////
namespace saga { namespace adaptors {

  /////////////////////////////////////////////////////////////////////////////
  //  this is the same as the saga::adaptor class, except it is possible to set
  //  read only attributes
  class SAGA_EXPORT attribute 
  :   public saga::detail::attribute<attribute>
  {
    public:
      // just defined for brevity
      typedef std::vector <std::string>              strvec_type;
      typedef std::map    <std::string, std::string> strmap_type;

    private:
      friend struct detail::attribute<attribute>;  // needs to access get_impl()
      
      // This members get initialized differently depending on the constructor 
      // used:
      //  - if this object got constructed from a saga::object, then the impl_ 
      //    member get's initialized directly with the corresponding 
      //    implementation object (saga::impl::object), and the attr_ member
      //    from is initialized the saga::impl::attribute corresponding to the 
      //    given impl_.
      //  - if this object got constructed from a saga::impl::v1_0::cpi*,
      //    we assume a cpi instance wants to get access to _its_own_
      //    attributes. We initialize the impl_ from the corresponding proxy,
      //    and the attr_ from the (default) attribute_cache_cpi implementation 
      //    our cpi type has been derived from.
      saga::impl::object* impl_;
      saga::impl::attribute_interface* attr_;
      bool must_delete;
      
      template <typename Tag> 
        task_type set_attributepriv (saga::void_t&, std::string const&, std::string const&, Tag);
      template <typename Tag> 
        task_type set_vector_attributepriv(saga::void_t&, std::string const&, strvec_type const&, Tag);
      template <typename Tag> 
        task_type remove_attributepriv(saga::void_t&, std::string const&, Tag);

    public:
      saga::impl::object* get_impl() const { return impl_; }
      saga::impl::attribute_interface* get_attr() const { return attr_; }

      // ctor/dtor
      explicit attribute (saga::impl::v1_0::attribute_cpi *impl);    // adaptor interface
      attribute (saga::object rhs);
      ~attribute (void);

      /*! \brief Brief %description starts here    
       *
       *   Detailed %description starts here.
       */  
      saga::void_t set_attribute(std::string key, std::string value) {
          saga::void_t ret;
          set_attributepriv(ret, key, value, saga::task_base::Sync());
          return ret;
      }
      SAGA_CALL_PUB_2_DEF_0(saga::void_t, set_attribute, std::string, std::string)

      /*! \brief Brief %description starts here    
       *
       *   Detailed %description starts here.
       */  
      saga::void_t set_vector_attribute(std::string key, strvec_type value) {
          saga::void_t ret;
          set_vector_attributepriv(ret, key, value, saga::task_base::Sync());
          return ret;
      }
      SAGA_CALL_PUB_2_DEF_0(saga::void_t, set_vector_attribute, std::string, strvec_type)
      
      /*! \brief Brief %description starts here    
       *
       *   Detailed %description starts here.
       */  
      saga::void_t remove_attribute(std::string key) {
          saga::void_t ret;
          remove_attributepriv(ret, key, saga::task_base::Sync());
          return ret;
      }
      SAGA_CALL_PUB_1_DEF_0(saga::void_t, remove_attribute, std::string)

  };

///////////////////////////////////////////////////////////////////////////////
}} // namespace saga::adaptors
                                                                             
#if defined(BOOST_MSVC)
#pragma warning(pop)
#endif

#endif // SAGA_ADAPTORS_ATTRIBUTE_HPP

