//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <boost/plugin/dll.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>

#include <saga/util.hpp>
#include <impl/runtime.hpp>
#include <saga/version.hpp>

#include <impl/engine/session.hpp>
#include <saga/adaptors/serialization.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace saga { namespace adaptors
{

// retrieve the name of the package containing for the implementation of the
// given SAGA object type
inline char const* const 
get_package_name(saga::object obj, saga::object::type type)
{
// The sequence of package names in this array should be equivalent to the 
// definition sequence of the corresponding saga object types 
// (see saga/object.hpp)
    static char const* const package_names[] = 
    {
        "saga",           // Exception        =   1,
        "saga",           // Session          =   2,
        "saga",           // Context          =   3,
        "saga",           // Task             =   4,
        "saga",           // TaskContainer    =   5,
        "saga",           // Buffer           =   6,
        "saga",           // Metric           =   7,
        "namespace",      // NSEntry          =   8,
        "namespace",      // NSDirectory      =   9,
        "file",           // IOVec            =  10,
        "file",           // File             =  11,
        "file",           // Directory        =  12,
        "logicalfile",    // LogicalFile      =  13,
        "logicalfile",    // LogicalDirectory =  14,
        "job",            // JobDescription   =  15,
        "job",            // JobService       =  16,
        "job",            // Job              =  17,
        "job",            // JobSelf          =  18,
        "stream",         // StreamService    =  19,
        "stream",         // Stream           =  20,
        "rpc",            // RPC              =  21,
        "rpc",            // Parameter        =  22,
            
        // SAGA Message API extension
        "message",        // Msg              =  23,
        "message",        // Endpoint         =  24,
            
        // SAGA Information services extension
        "advert",         // Advert           =  25,
        "advert",         // AdvertDirectory  =  26
    };

    if (type < saga::object::Exception || type > saga::object::AdvertDirectory)
    {
        SAGA_THROW_VERBATIM(obj, "Attempt to retrieve package name from "
            "unknown SAGA object type.", saga::BadParameter);
    }
    return package_names[type-1];
}

// serialize a saga::object into a string representation
std::string serialize(saga::object obj)
{
    using namespace boost::serialization;
    using namespace boost::archive;
    
    SAGA_OSSTREAM strm;
    {
        try {
          text_oarchive oa(strm);
          
          // save the version of the SAGA engine
          unsigned int version = SAGA_VERSION_FULL;
          oa << make_nvp("saga_version", version);
          
          // save the object type
          saga::object::type type = obj.get_type();
          oa << make_nvp("type", type);
          
          saga::uuid objid (obj.get_id());
          const std::string str(objid.string());
          oa << make_nvp("id", str);
          
          // save the object implementation
          TR1::shared_ptr<saga::impl::object> impl(
              saga::impl::runtime::get_impl_object(obj));
        
          // load the corresponding package and delegate the serialization
          std::string basename (get_package_name (obj, type));

#ifdef BOOST_WINDOWS
#if defined(_DEBUG)
          std::string package (               basename + "d" + SAGA_SHARED_LIB_EXTENSION);
#else
          std::string package (                     basename + SAGA_SHARED_LIB_EXTENSION);
#endif
#else
          std::string package ("libsaga_package_" + basename + SAGA_SHARED_LIB_EXTENSION);
#endif

          boost::plugin::dll d (package);
          boost::plugin::plugin_factory <saga::adaptors::serialization> pf (d);

          TR1::shared_ptr <saga::adaptors::serialization> ser (pf.create ("serialization")); 

          // serialize the object
          std::string data (ser->serialize(impl));
        
          oa << make_nvp ("object", data);
        }
        catch (archive_exception const& e) {
            SAGA_THROW_VERBATIM(obj, "Serialization failed (archive exception): " + 
                e.what(), saga::NoSuccess);
        }
        catch (std::logic_error const& e) {
            SAGA_THROW_VERBATIM(obj, "Serialization failed: " + e.what(), 
                saga::NoSuccess);
        }
    } 
    return SAGA_OSSTREAM_GETSTRING(strm);
}

// create a saga::object from a given string representation
saga::object deserialize(std::string const& data)
{
    return deserialize(saga::detail::get_the_session(), data);
}

saga::object deserialize(saga::session s, std::string const& data)
{
    using namespace boost::serialization;
    using namespace boost::archive;
    
    TR1::shared_ptr<saga::impl::object> impl;
    std::string objid;
    
    try {
      SAGA_ISSTREAM strm(data);
      text_iarchive ia(strm);

      // retrieve the version of the SAGA engine and make sure it is compatible
      // with our current version
      unsigned int version = 0;
      ia >> make_nvp("saga_version", version);  // load version
      if (!SAGA_VERSION_ISCOMPATIBLE_EX(version)) 
      {
          SAGA_THROW_VERBATIM(s, "Deserialization failed: "
              "incompatible version of SAGA modules.", saga::NoSuccess);
      }
      
      // retrieve the object type of the stored object
      saga::object::type type;
      ia >> make_nvp("type", type);
      ia >> make_nvp("id", objid);
      
      // load the package corresponding to the object type
      std::string basename (get_package_name (s, type));

#ifdef BOOST_WINDOWS
#if defined(_DEBUG)
      std::string package (               basename + "d" + SAGA_SHARED_LIB_EXTENSION);
#else
      std::string package (                     basename + SAGA_SHARED_LIB_EXTENSION);
#endif
#else
      std::string package ("libsaga_package_" + basename + SAGA_SHARED_LIB_EXTENSION);
#endif

      boost::plugin::dll d(package);
      boost::plugin::plugin_factory <saga::adaptors::serialization> pf (d);
      TR1::shared_ptr <saga::adaptors::serialization> ser (pf.create("serialization")); 

      // deserialize the given string
      std::string data;
      ia >> make_nvp("object", data);
      impl = ser->deserialize(s, type, data);
    }
    catch (archive_exception const& e) {
        SAGA_THROW_VERBATIM(s, "Deserialization failed (archive exception): " + 
            e.what(), saga::NoSuccess);
    }
    catch (std::logic_error const& e) {
        SAGA_THROW_VERBATIM(s, "Deserialization failed: " + e.what(),
            saga::NoSuccess);
    }
    
    // set the objid
    impl->set_uuid(saga::uuid(objid.c_str()));
    return saga::impl::runtime::get_object(impl);
}

///////////////////////////////////////////////////////////////////////////////
}}    // namespace saga::adaptors
