//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Use, modification and distribution is subject to the Boost Software
//  License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)

#include <boost/config.hpp>
#include <boost/lexical_cast.hpp>   // lexical_cast will be in TR2

#include <saga/exception.hpp>
#include <saga/context.hpp>

#include <impl/context.hpp>
#include <impl/engine/object.hpp>
#include <impl/engine/metric.hpp>

#include <saga/detail/attribute_impl.hpp>

#include <boost/assign/std/vector.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace saga 
{
  namespace detail
  {
    ///@cond
      std::string get_userid()
      {
#if defined(BOOST_WINDOWS)
          char chBuffer[MAX_PATH + 1];
          DWORD cchBuff = sizeof(chBuffer)-1;
          GetUserName (chBuffer, &cchBuff);     // Get and display the user name.
          return chBuffer;
#else
          return boost::lexical_cast<std::string>(getuid());
#endif
      }
    ///@endcond
  }
  
  /////////////////////////////////////////////////////////////////////////////
  context::context (saga::context::type t) 
    : saga::object (new impl::context(t))
  {
    using namespace boost::assign;
    
    strmap_type attributes_scalar_rw;
    strmap_type attributes_scalar_ro;
    
    insert(attributes_scalar_ro)
        (attributes::context_remote_id, "")
        (attributes::context_remote_host, "")
        (attributes::context_remote_port, "");
        
    switch (t) {
    case context_base::X509:
        {
            std::vector<std::string> valid_keys;
            valid_keys += 
                attributes::context_remote_host,
                attributes::context_remote_id, 
                attributes::context_remote_port,
                attributes::context_cert_dir,
                attributes::context_user_cert,
                attributes::context_user_proxy,
                attributes::context_user_key,
                attributes::context_user_pass;

            this->init_keynames(valid_keys);
            insert(attributes_scalar_rw)
                (attributes::context_user_proxy, std::string("/tmp/x509up_u") + 
                    detail::get_userid())
                (attributes::context_user_cert, std::string("$HOME/.globus/usercert.pem"))
                (attributes::context_user_key, std::string("$HOME/.globus/userkey.pem"))
                (attributes::context_user_pass, "")
                (attributes::context_cert_dir, "/etc/grid-security/certificates/");
            this->init (attributes_scalar_ro, attributes_scalar_rw);
        }
        break;
        
    case context_base::MyProxy:
        {
            std::vector<std::string> valid_keys;
            valid_keys += 
                attributes::context_server,
                attributes::context_remote_host,
                attributes::context_remote_id, 
                attributes::context_remote_port,
                attributes::context_cert_dir,
                attributes::context_user_cert,
                attributes::context_user_proxy,
                attributes::context_user_key,
                attributes::context_user_pass;

            this->init_keynames(valid_keys);
            insert(attributes_scalar_rw)
                (attributes::context_server, "localhost")
                (attributes::context_user_proxy, std::string("/tmp/x509up_u") + 
                    detail::get_userid())
                (attributes::context_user_cert, "$HOME/.globus/usercert.pem")
                (attributes::context_user_key, "$HOME/.globus/userkey.pem")
                (attributes::context_user_pass, "")
                (attributes::context_cert_dir, "/etc/grid-security/certificates/");
            this->init (attributes_scalar_ro, attributes_scalar_rw);
        }
        break;
        
    case context_base::SAML:
        {
            std::vector<std::string> valid_keys;
            valid_keys += 
                attributes::context_remote_host,
                attributes::context_remote_id, 
                attributes::context_remote_port,
                attributes::context_user_name,
                attributes::context_user_pass;

            this->init_keynames(valid_keys);
            insert(attributes_scalar_rw)
                (attributes::context_user_name, "")
                (attributes::context_user_pass, "");
            this->init (attributes_scalar_ro, attributes_scalar_rw);
        }
        break;
        
    case context_base::Unicore:
        {
            std::vector<std::string> valid_keys;
            valid_keys += 
                attributes::context_remote_host,
                attributes::context_remote_id, 
                attributes::context_remote_port,
                attributes::context_user_cert,
                attributes::context_user_pass;

            this->init_keynames(valid_keys);
            insert(attributes_scalar_rw)
                (attributes::context_user_cert, "$HOME/.keystore")
                (attributes::context_user_pass, "");
            this->init (attributes_scalar_ro, attributes_scalar_rw);
        }
        break;
        
    case context_base::SSH:
        {
            std::vector<std::string> valid_keys;
            valid_keys += 
                attributes::context_remote_host,
                attributes::context_remote_id, 
                attributes::context_remote_port,
                attributes::context_cert_dir,
                attributes::context_user_cert,
                attributes::context_user_pass;

            this->init_keynames(valid_keys);
            insert(attributes_scalar_rw)
                (attributes::context_cert_dir, "$HOME/.ssh/")
                (attributes::context_user_cert, "$HOME/.ssh/id_dsa.pub")
                (attributes::context_user_pass, "$HOME/.ssh/id_dsa");
            this->init (attributes_scalar_ro, attributes_scalar_rw);
        }
        break;
        
    case context_base::Kerberos:
        {
            std::vector<std::string> valid_keys;
            valid_keys += 
                attributes::context_server,
                attributes::context_remote_host,
                attributes::context_remote_id, 
                attributes::context_remote_port,
                attributes::context_user_name,
                attributes::context_user_pass;

            this->init_keynames(valid_keys);
            insert(attributes_scalar_rw)
                (attributes::context_server, "localhost")
                (attributes::context_user_name, "")
                (attributes::context_user_pass, "");
            this->init (attributes_scalar_ro, attributes_scalar_rw);
        }
        break;
        
    case context_base::UserPass:
        {
            std::vector<std::string> valid_keys;
            valid_keys += 
                attributes::context_remote_host,
                attributes::context_remote_id, 
                attributes::context_remote_port,
                attributes::context_user_name,
                attributes::context_user_pass;

            this->init_keynames(valid_keys);
            insert(attributes_scalar_rw)
                (attributes::context_user_name, "anonymous")
                (attributes::context_user_pass, "anon");
            this->init (attributes_scalar_ro, attributes_scalar_rw);
        }
        break;

    case context_base::Unknown:
        break;    // no supported attributes
        
    default:
        SAGA_THROW("Invalid context type.", saga::BadParameter);
    }
    
    // initialize attribute implementation
    this->init (false, true);   // attribute_cache only
  }
  
  context::context(saga::impl::context *impl)
    : saga::object (impl)
  {
  }
  
  context::context(saga::object const& obj)
    : saga::object (obj)
  {
      if (this->get_type() != saga::object::Context)
      {
          SAGA_THROW("Bad type conversion.", saga::BadParameter);
      }
  }
  context& context::operator=(saga::object const& obj)
  {
      saga::object::operator=(obj);
      if (this->get_type() != saga::object::Context)
      {
          SAGA_THROW("Bad type conversion.", saga::BadParameter);
      }
      return *this;
  }
  
  context::~context (void)
  {
  }
  
  TR1::shared_ptr <saga::impl::context> context::get_impl (void) const
  { 
    typedef saga::object base_type;
    return TR1::static_pointer_cast<saga::impl::context>(
        this->base_type::get_impl());
  }

  /// @cond 
  /** Doxygen generates namespace function from friend impls... */
  bool operator== (context const & lhs, context const & rhs)
  {
      if (lhs.get_ctype() == saga::context_base::Unknown ||
          rhs.get_ctype() == saga::context_base::Unknown)
      {
          SAGA_THROW_VERBATIM(lhs, "Can't compare a context of type 'Unknown'.", 
              saga::IncorrectState);
      }
      return lhs.get_impl()->is_equal(rhs.get_impl().get()); 
  }
  /// @endcond

  context::type context::get_ctype (void) const
  {
    return get_impl()->get_ctype ();
  }
  
  /////////////////////////////////////////////////////////////////////////////
  //  implement the attribute functions (we need to explicitly specialize 
  //  the template because the functions are not implemented inline)
  template struct detail::attribute<context>;
  
  template struct detail::attribute_priv<context, task_base::Sync>;
  template struct detail::attribute_priv<context, task_base::Async>;
  template struct detail::attribute_priv<context, task_base::Task>;
  
///////////////////////////////////////////////////////////////////////////////
} // namespace saga 

