//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005 Michel Zandstra (michelzandstra@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef SAGA_IMPL_ENGINE_BASE_HPP
#define SAGA_IMPL_ENGINE_BASE_HPP

#include <cstdlib>
#include <cstddef>

#include <saga/export_definitions.hpp>

///////////////////////////////////////////////////////////////////////////////
#define SAGA_CPIVERSION_1_0       0x0100
#define SAGA_CPIVERSION_MINORMASK 0x000F
#define SAGA_CPIVERSION_LATEST    SAGA_CPIVERSION_1_0

///////////////////////////////////////////////////////////////////////////////

/// @cond
/** Hide from Doxygen */
namespace saga
{
  namespace adaptors 
  {
    class task;
    class attribute;
    class metric;
    class exception;
  }

  // forward declaration for implementation
  namespace impl
  {
    struct runtime;
    class url;
    class uuid;
    class const_buffer;
    class buffer;
    
    class object;
    class session;
    class context;
    class attribute;
    class permissions;
    struct permissions_interface;
    
    struct task_interface;
    class task_base;
    class task_container;

    class namespace_entry;
    class namespace_dir;

    class directory;
    class file;
    class const_iovec;
    class iovec;

    class logical_directory;
    class logical_file;

    class advert;
    class advert_directory;

    class rpc;
    class parameter;
    
    class stream;
    class server;

    class job;
    class self;
    class service;
    class description;
    class istream_interface;
    class ostream_interface;
    
//  class cpr;

    class metric;
//  class monitor;
//  class actor;

    class adaptor_selector;

    class proxy;

    namespace v1_0 
    {
      class cpi;
      class cpi_info;
      class op_info;
    }
    
    // interfaces
    class attribute;
    class monitorable;
    class steerable;
    struct task_interface;
    
    // exception handling
    class exception_list;
    
  } // namespace impl

    ///////////////////////////////////////////////////////////////////////////
    // define a thread safe getenv() function (use this instead of direct calls
    // to getenv
    
    /// @cond
    /** These methods are not within API scope TODO: wrong namespace?!? */
    SAGA_EXPORT char* safe_getenv(const char *);
    /// @endcond

} // namespace saga

/// @endcond

///////////////////////////////////////////////////////////////////////////////
//
// Macros for verbose and debug output
//

#ifdef SAGA_LOG_FACILITY
#  define SAGA_VERBOSE_LOG_FACILITY SAGA_LOG_FACILITY
#else
#  define SAGA_VERBOSE_LOG_FACILITY __FILE__
#endif

///////////////////////////////////////////////////////////////////////////////
#define SAGA_VERBOSE(x)                                                       \
        if ( NULL !=       saga::safe_getenv ("SAGA_VERBOSE")  &&             \
                x <= atoi (saga::safe_getenv ("SAGA_VERBOSE")) )              \
  /**/

#define SAGA_LOG(level,msg)                                                   \
       SAGA_VERBOSE (level)                                                   \
       {                                                                      \
          fprintf (stderr, "%-10s : %-20s : %s\n",                            \
                   level ## _MSG, SAGA_VERBOSE_LOG_FACILITY, msg);            \
       }                                                                      \
  /**/

#define SAGA_LOG_CRITICAL(msg)                                                \
        SAGA_LOG (SAGA_VERBOSE_LEVEL_CRITICAL,msg)                            \
  /**/

#define SAGA_LOG_ERROR(msg)                                                   \
        SAGA_LOG (SAGA_VERBOSE_LEVEL_ERROR,msg)                               \
  /**/

#define SAGA_LOG_WARN(msg)                                                    \
        SAGA_LOG (SAGA_VERBOSE_LEVEL_WARNING,msg)                             \
  /**/

#define SAGA_LOG_INFO(msg)                                                    \
        SAGA_LOG (SAGA_VERBOSE_LEVEL_INFO,msg)                                \
  /**/

#define SAGA_LOG_DEBUG(msg)                                                   \
        SAGA_LOG (SAGA_VERBOSE_LEVEL_DEBUG,msg)                               \
  /**/

#define SAGA_LOG_BLURB(msg)                                                   \
        SAGA_LOG (SAGA_VERBOSE_LEVEL_BLURB,msg)                               \
  /**/

char const* const SAGA_VERBOSE_LEVEL_CRITICAL_MSG      = "CRITICAL";
char const* const SAGA_VERBOSE_LEVEL_ERROR_MSG         = "ERROR";
char const* const SAGA_VERBOSE_LEVEL_WARNING_MSG       = "WARNING";
char const* const SAGA_VERBOSE_LEVEL_INFO_MSG          = "INFO";
char const* const SAGA_VERBOSE_LEVEL_DEBUG_MSG         = "DEBUG";
char const* const SAGA_VERBOSE_LEVEL_BLURB_MSG         = "BLURB";

#define SAGA_VERBOSE_LEVEL_CRITICAL       1
#define SAGA_VERBOSE_LEVEL_ERROR          2
#define SAGA_VERBOSE_LEVEL_WARNING        3
#define SAGA_VERBOSE_LEVEL_INFO           4
#define SAGA_VERBOSE_LEVEL_DEBUG          5
#define SAGA_VERBOSE_LEVEL_BLURB          6

///////////////////////////////////////////////////////////////////////////////
// Helper macros/functions to overcome the lack of stringstream on certain 
// architectures
#ifdef BOOST_NO_STRINGSTREAM
# include <strstream>
  inline std::string SAGA_OSSTREAM_GETSTRING (std::ostrstream & ss)
  {
    ss << std::ends;
    std::string rval = ss.str ();
    ss.freeze (false);
    return (rval);
  }
# define SAGA_OSSTREAM std::ostrstream
# define SAGA_ISSTREAM std::istrstream
#else
# include <sstream>
# define SAGA_OSSTREAM_GETSTRING(ss) ss.str()
# define SAGA_OSSTREAM std::ostringstream
# define SAGA_ISSTREAM std::istringstream
#endif

///////////////////////////////////////////////////////////////////////////////
#if defined(__AIX__) && defined(__GNUC__)
// we must call the initialization functions of the shared libraries ourselves, 
// because AIX's dlopen() system call does not know about it. 
extern "C" {
  void _GLOBAL__DI();   // call constructors of all global objects
  void _GLOBAL__DD();   // call destructors of all global objects
}
#endif

#endif // SAGA_IMPL_ENGINE_BASE_HPP


