#if defined(__WAVE__) && defined(SAGA_CREATE_PREPROCESSED_FILES)
#pragma wave option(preserve: 2, line: 1, output: "preprocessed/job.cpp")
#endif
//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if defined(__WAVE__) && defined(SAGA_CREATE_PREPROCESSED_FILES)
#pragma wave option(output: null)
#endif

//  this is needed in every file including detail/attribute_impl.hpp and not 
//  belonging to the engine
#define SAGA_NO_EXPORT_ATTRIBUTE
#define SAGA_NO_EXPORT_PERMISSIONS

#include <boost/assign/list_inserter.hpp>
#include <boost/assign/std/vector.hpp>

// include job API and implementation
#include <saga/job.hpp>
#include <saga/adaptors/metric.hpp>
#include <impl/job.hpp>

#include <saga/detail/attribute_impl.hpp>
#include <saga/detail/permissions_impl.hpp>

#ifdef SAGA_DEBUG
#include <saga/packages/job/preprocessed/job.cpp>
#else

#if defined(__WAVE__) && defined(SAGA_CREATE_PREPROCESSED_FILES)
#pragma wave option(preserve: 2, line: 1, output: "preprocessed/job.cpp")
#endif
///////////////////////////////////////////////////////////////////////////////
namespace saga { namespace job {

    namespace metrics
    {
        /// @cond
        saga::metrics::init_data const job_metric_data[] = 
        {
          // task specific metrics
          {
              saga::metrics::task_state,
              "Metric to monitor the current state of the job, "
                  "e.g. New, Running, Canceled, Done, or Failed",
              saga::attributes::metric_mode_readonly,
              "1", 
              saga::attributes::metric_type_enum,
              "0"   // saga::job::New
          },
          
          // job specific metrics
          {
              saga::job::metrics::statedetail,
              "Metric fires on state detail changes of the job.",
              saga::attributes::metric_mode_readonly,
              "1", 
              saga::attributes::metric_type_string,
              ""
          },
          {
              saga::job::metrics::signal,
              "Metric fires as a job receives a signal, and has a "
                  "value indicating the signal number.",
              saga::attributes::metric_mode_readonly,
              "1", 
              saga::attributes::metric_type_int,
              ""
          },
          {
              saga::job::metrics::cputimelimit,
              "Metric gives access to number of cpu seconds consumed by the job.",
              saga::attributes::metric_mode_readonly,
              "seconds", 
              saga::attributes::metric_type_int,
              ""
          },
          {
              saga::job::metrics::memoryuse,
              "Metric gives access to the current aggregate memory usage "
                  "of the job.",
              saga::attributes::metric_mode_readonly,
              "megabyte", 
              saga::attributes::metric_type_float,
              "0.0"
          },
          {
              saga::job::metrics::vmemoryuse,
              "Metric gives access to the current aggregate virtual memory "
                  "usage of the job.",
              saga::attributes::metric_mode_readonly,
              "megabyte", 
              saga::attributes::metric_type_float,
              "0.0"
          },
          {
              saga::job::metrics::performance,
              "Metric gives access to the current performance of the job.",
              saga::attributes::metric_mode_readonly,
              "FLOPS", 
              saga::attributes::metric_type_float,
              "0.0"
          },
      };
      /// @endcond
      
}
  

  
  /////////////////////////////////////////////////////////////////////////////
  job::job()
  {
  }
  
  job::job(saga::impl::job* impl)
    : task(impl)
  {
      // initialize attributes
      using namespace boost::assign;
      std::vector<std::string> valid_keys;
      valid_keys += 
              attributes::jobid,
              attributes::executionhosts,
              attributes::created,
              attributes::started,
              attributes::finished,
              attributes::workingdirectory,
              attributes::exitcode,
              attributes::termsig
          ;

      // initialize list of valid keys          
      this->attribute_base::init_keynames(valid_keys);

      strmap_type attributes_scalar_ro;
      insert(attributes_scalar_ro)
              (attributes::jobid, "")
          ;

      // initialize attribute implementation
      this->attribute_base::init (attributes_scalar_ro);
      
      // initialize metrics
      std::vector<saga::metric> metrics;
      for (unsigned int i = 0; 
           i < sizeof(saga::job::metrics::job_metric_data)/sizeof(saga::metrics::init_data);
           ++i)
      {
          saga::metrics::init_data const* p = &saga::job::metrics::job_metric_data[i];
          saga::metric m(*this, p->name, p->description, p->mode, p->unit, 
              p->type, p->value);
          metrics.push_back(m);
      }
      this->monitorable_base::init (metrics);

      this->saga::object::get_impl()->init();
  }
  
  job::~job (void)
  {
  }

  /////////////////////////////////////////////////////////////////////////////
  job::job (saga::object const& o)
    : saga::task_type (o)
  {
      if (this->get_type() != saga::object::Job)
      {
          SAGA_THROW("Bad type conversion.", saga::BadParameter);
      }
  }

  job &job::operator= (saga::object const& o)
  {
      return saga::task_type::operator=(o), *this;
  }

  /////////////////////////////////////////////////////////////////////////////
  TR1::shared_ptr <saga::impl::job> job::get_impl (void) const
  { 
    // FIXME: this needs documentation
    typedef saga::object base_type;
    return TR1::static_pointer_cast <saga::impl::job> (
        this->base_type::get_impl ());
  }

  // job inspection
  SAGA_CALL_IMP_0 (job, std::string, get_job_id)
  SAGA_CALL_IMP_0 (job, saga::job::state, get_state)
  SAGA_CALL_IMP_0 (job, description, get_description)
  SAGA_CALL_IMP_0 (job, saga::job::ostream, get_stdin)
  SAGA_CALL_IMP_0 (job, saga::job::istream, get_stdout)
  SAGA_CALL_IMP_0 (job, saga::job::istream, get_stderr)

  // job management
  SAGA_CALL_IMP_0 (job, saga::void_t, suspend)
  SAGA_CALL_IMP_0 (job, saga::void_t, resume)
  SAGA_CALL_IMP_0 (job, saga::void_t, checkpoint)
  SAGA_CALL_IMP_1 (job, saga::void_t, migrate, description)
  SAGA_CALL_IMP_1 (job, saga::void_t, signal, int)

  }   // namspace job
  
  /////////////////////////////////////////////////////////////////////////////
  //  implement the attribute functions (we need to explicitly specialize 
  //  the template because the functions are not implemented inline)
  template struct detail::attribute<job::job>;
  
  template struct detail::attribute_priv<job::job, task_base::Sync>;
  template struct detail::attribute_priv<job::job, task_base::Async>;
  template struct detail::attribute_priv<job::job, task_base::Task>;
  
  /////////////////////////////////////////////////////////////////////////////
  //  implement the permissions functions (we need to explicitly specialize 
  //  the template because the functions are not implemented inline)
  template struct detail::permissions<job::job>;
  
  template struct detail::permissions_priv<job::job, task_base::Sync>;
  template struct detail::permissions_priv<job::job, task_base::Async>;
  template struct detail::permissions_priv<job::job, task_base::Task>;
  
///////////////////////////////////////////////////////////////////////////////////
} // namespace saga

#endif

